# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('GRID')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.grid["inputs"].keys():
    defaults[key] = schema.grid["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.grid)
def grid(**kw):
    '''
    Compute imaging weights and create a dirty image, psf etc.
    By default only the MFS images are converted to fits files.
    Set the --fits-cubes flag to also produce fits cubes.

    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'grid_{timestamp}.log')

    if opts.nworkers is None:
        if opts.scheduler=='distributed':
            opts.nworkers = opts.nband
        else:
            opts.nworkers = 1

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _grid(**opts)

def _grid(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import os
    import numpy as np
    import dask
    dask.config.set(**{'array.slicing.split_large_chunks': False})
    from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
    import dask.array as da
    from africanus.constants import c as lightspeed
    from ducc0.fft import good_size
    from pfb.utils.fits import dds2fits, dds2fits_mfs
    from pfb.utils.misc import compute_context
    from pfb.operators.gridder import vis2im, loc2psf_vis
    from pfb.operators.fft import fft2d
    from pfb.utils.weighting import (compute_counts, counts_to_weights,
                                     filter_extreme_counts)
    from pfb.utils.beam import eval_beam
    import xarray as xr
    from uuid import uuid4
    from daskms.optimisation import inlined_array
    from pfb.utils.astrometry import get_coordinates
    from africanus.coordinates import radec_to_lm

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    # xds contains vis products, no imaging weights applied
    xds_name = f'{basename}.xds'
    xdsp = xds_from_zarr(xds_name, chunks={'row': -1, 'chan': -1})
    # dds contains image space products including imaging weights and uvw
    dds_name = f'{basename}_{opts.postfix}.dds'

    if os.path.isdir(dds_name):
        dds_exists = True
        if opts.overwrite:
            print(f'Removing {dds_name}', file=log)
            import shutil
            shutil.rmtree(dds_name)
            dds_exists = False
    else:
        dds_exists = False

    if opts.concat and len(xdsp) > opts.nband:
        # this is required because concat will try to mush different
        # imaging bands together if they are not split upfront
        print('Concatenating datasets along row dimension', file=log)
        xds = []
        for b in range(opts.nband):
            xdsb = []
            times = []
            timeids = []
            for ds in xdsp:
                if ds.bandid == b:
                    xdsb.append(ds)
                    times.append(ds.time_out)
                    timeids.append(ds.timeid)
            xdso = xr.concat(xdsb, dim='row',
                             data_vars='minimal',
                             coords='minimal').chunk({'row':-1})
            tid = np.array(timeids).min()
            tout = np.mean(np.array(times))
            xdso.assign_attrs(
                {'time_out': tout, 'timeid': tid}
            )
            xds.append(xdso)
        try:
            assert len(xds) == opts.nband
        except Exception as e:
            raise RuntimeError('Something went wrong during concatenation.'
                               'This is probably a bug.')
        ntime = 1
    else:
        xds = xdsp
        ntime = len(xds) // opts.nband
        if len(xds) < opts.nband:
            raise RuntimeError("len(xds) < nband. Run with the same nband as init worker.")
        elif len(xds) % ntime:
            raise RuntimeError("len(xds) != ntime*nband. This is a bug.")

    real_type = xds[0].WEIGHT.dtype
    if real_type == np.float32:
        precision = 'single'
    else:
        precision = 'double'

    # max uv coords over all datasets
    uv_maxs = []
    max_freqs = []
    for ds in xds:
        uvw = ds.UVW.data
        u_max = abs(uvw[:, 0]).max()
        v_max = abs(uvw[:, 1]).max()
        uv_maxs.append(da.maximum(u_max, v_max))
        max_freqs.append(ds.FREQ.data.max())

    # early compute necessary to set image size
    uv_maxs, max_freqs = dask.compute(uv_maxs, max_freqs)
    uv_max = max(uv_maxs)
    max_freq = max(max_freqs)

    # max cell size
    cell_N = 1.0 / (2 * uv_max * max_freq / lightspeed)

    if opts.cell_size is not None:
        cell_size = opts.cell_size
        cell_rad = cell_size * np.pi / 60 / 60 / 180
        if cell_N / cell_rad < 1:
            raise ValueError("Requested cell size too small. "
                             "Super resolution factor = ", cell_N / cell_rad)
        print(f"Super resolution factor = {cell_N/cell_rad}", file=log)
    else:
        cell_rad = cell_N / opts.super_resolution_factor
        cell_size = cell_rad * 60 * 60 * 180 / np.pi
        print(f"Cell size set to {cell_size} arcseconds", file=log)

    if opts.nx is None:
        fov = opts.field_of_view * 3600
        npix = int(fov / cell_size)
        npix = good_size(npix)
        while npix % 2:
            npix += 1
            npix = good_size(npix)
        nx = npix
        ny = npix
    else:
        nx = opts.nx
        ny = opts.ny if opts.ny is not None else nx

    nband = opts.nband
    if opts.dirty:
        print(f"Image size = ({nband}, {nx}, {ny})", file=log)

    nx_psf = good_size(int(opts.psf_oversize * nx))
    while nx_psf % 2:
        nx_psf += 1
        nx_psf = good_size(nx_psf)

    ny_psf = good_size(int(opts.psf_oversize * ny))
    while ny_psf % 2:
        ny_psf += 1
        ny_psf = good_size(ny_psf)

    if opts.psf:
        print(f"PSF size = ({nband}, {nx_psf}, {ny_psf})", file=log)

    # print(f'Not overwriting {dds_name}, directory exists. '
    #               f'Set overwrite flag if you mean to overwrite.', file=log)
    # if dds exists, check that existing dds is compatible with input
    if dds_exists:
        dds = xds_from_zarr(dds_name)
        # these need to be aligned at this stage
        for ds, out_ds in zip(xds, dds):
            assert ds.bandid == out_ds.bandid
            assert ds.freq_out == out_ds.freq_out
            assert ds.timeid == out_ds.timeid
            assert ds.time_out == out_ds.time_out
            assert ds.ra == out_ds.ra
            assert ds.dec == out_ds.dec
            assert out_ds.x.size == nx
            assert out_ds.y.size == ny
            assert out_ds.cell_rad == cell_rad
        print(f'As far as we can tell {dds_name} can be reused/updated.',
              file=log)
    else:
        print(f'Image space data products will be stored in {dds_name}.', file=log)

    # check if model exists
    if opts.transfer_model_from is not None:
        try:
            mds = xds_from_zarr(opts.transfer_model_from,
                                chunks={'x':-1, 'y':-1})
        except Exception as e:
            raise ValueError(f"No dataset found at {opts.transfer_model_from}")
        try:
            assert len(mds) == len(xds)
            for ms, ds in zip(mds, xds):
                assert ms.bandid == ds.bandid
        except Exception as e:
            raise ValueError("Transfer from dataset mismatched. "
                             "This is not currently supported.")
        try:
            assert 'MODEL' in mds[0]
        except Exception as e:
            raise ValueError(f"No MODEL variable in {opts.transfer_model_from}")

        print(f"Found MODEL in {opts.transfer_model_from}. ",
              file=log)
        has_model = True
    else:
        has_model = False

    dds_out = []
    for i, ds in enumerate(xds):
        if dds_exists:
            out_ds = dds[i]
        else:
            out_ds = xr.Dataset()
        uvw = ds.UVW.data
        freq = ds.FREQ.data
        vis = ds.VIS.data
        wgt = ds.WEIGHT.data
        # This is a vis space mask (see wgridder convention)
        mask = ds.MASK.data
        if opts.robustness is not None:
            # we'll skip this if counts already exists
            # what to do if flags have changed?
            if 'COUNTS' not in out_ds:
                counts = compute_counts(
                        uvw,
                        freq,
                        mask,
                        nx,
                        ny,
                        cell_rad,
                        cell_rad,
                        wgt.dtype)
                # get rid of artificially high weights corresponding to
                # nearly empty cells
                if opts.filter_extreme_counts:
                    counts = filter_extreme_counts(counts, nbox=opts.filter_nbox,
                                                   nlevel=opts.filter_level)

                # do we want the coordinates to be ug, vg rather?
                out_ds = ds.assign(**{'COUNTS': (('x', 'y'), counts)})

            # we usually want to re-evaluate this since the robustness may change
            imwgt = counts_to_weights(out_ds.COUNTS.data,
                                      uvw,
                                      freq,
                                      nx, ny,
                                      cell_rad, cell_rad,
                                      opts.robustness)

            wgt *= imwgt


        # compute lm coordinates of target
        if opts.target is not None:
            obs_time = ds.time_out
            tra, tdec = get_coordinates(obs_time)
            tcoords=np.zeros((1,2))
            tcoords[0,0] = tra
            tcoords[0,1] = tdec
            coords0 = np.array((ds.ra, ds.dec))
            lm0 = radec_to_lm(tcoords, coords0).squeeze()
            # LB - why the negative?
            x0 = -lm0[0]
            y0 = -lm0[1]
        else:
            x0 = 0.0
            y0 = 0.0
            tra = ds.ra
            tdec = ds.dec

        # TODO - assign coordinates
        if not dds_exists:
            out_ds = out_ds.assign_attrs(**{
                'ra': tra,
                'dec': tdec,
                'x0': x0,
                'y0': y0,
                'cell_rad': cell_rad,
                'bandid': ds.bandid,
                'timeid': ds.timeid,
                'freq_out': ds.freq_out,
                'time_out': ds.time_out,
            })

        if opts.dirty:
            dirty = vis2im(uvw=uvw,
                           freq=freq,
                           vis=vis,
                           wgt=wgt,
                           nx=nx,
                           ny=ny,
                           cellx=cell_rad,
                           celly=cell_rad,
                           x0=x0, y0=y0,
                           nthreads=opts.nvthreads,
                           epsilon=opts.epsilon,
                           precision=precision,
                           mask=mask,
                           do_wgridding=opts.wstack,
                           double_precision_accumulation=opts.double_accum)
            dirty = inlined_array(dirty, [uvw, freq])
            out_ds = out_ds.assign(**{'DIRTY': (('x', 'y'), dirty)})

        if opts.psf:
            psf_vis = loc2psf_vis(uvw,
                                  freq,
                                  cell_rad,
                                  x0,
                                  y0,
                                  wstack=opts.wstack,
                                  epsilon=opts.epsilon,
                                  nthreads=opts.nvthreads,
                                  precision=precision)
            psf = vis2im(uvw=uvw,
                         freq=freq,
                         vis=psf_vis,
                         wgt=wgt,
                         nx=nx_psf,
                         ny=ny_psf,
                         cellx=cell_rad,
                         celly=cell_rad,
                         x0=x0, y0=y0,
                         nthreads=opts.nvthreads,
                         epsilon=opts.epsilon,
                         precision=precision,
                         mask=mask,
                         do_wgridding=opts.wstack,
                         double_precision_accumulation=opts.double_accum)
            psf = inlined_array(psf, [uvw, freq])
            # get FT of psf
            psfhat = fft2d(psf, nthreads=opts.nvthreads)
            out_ds = out_ds.assign(**{'PSF': (('x_psf', 'y_psf'), psf),
                                      'PSFHAT': (('x_psf', 'yo2'), psfhat)})

        # TODO - don't put vis space products in dds
        if opts.weight:
            out_ds = out_ds.assign(**{'WEIGHT': (('row', 'chan'), wgt)})

        wsum = wgt[mask.astype(bool)].sum()

        # evaluate beam at x and y coords
        cell_deg = np.rad2deg(cell_rad)
        l = (-(nx//2) + da.arange(nx)) * cell_deg + np.deg2rad(x0)
        m = (-(ny//2) + da.arange(ny)) * cell_deg + np.deg2rad(y0)
        ll, mm = da.meshgrid(l, m, indexing='ij')
        bvals = eval_beam(ds.BEAM.data, ll, mm)

        if opts.residual and 'MODEL' in out_ds:
            model = out_ds.MODEL.data
            from pfb.operators.hessian import hessian
            hessopts = {
                'cell': cell_rad,
                'wstack': opts.wstack,
                'epsilon': opts.epsilon,
                'double_accum': opts.double_accum,
                'nthreads': opts.nvthreads
            }
            # we only want to apply the beam once here
            residual = dirty - hessian(bvals * model, uvw, wgt,
                                       mask, freq, None, hessopts)
            residual = inlined_array(residual, [uvw, freq])
            out_ds = out_ds.assign(**{'RESIDUAL': (('x', 'y'), residual)})


        out_ds = out_ds.assign(**{'FREQ': (('chan',), freq),
                                  'UVW': (('row', 'three'), uvw),
                                  'MASK': (('row', 'chan'), mask),
                                  'WSUM': (('scalar',), da.atleast_1d(wsum)),
                                  'BEAM': (('x', 'y'), bvals)})



        out_ds = out_ds.chunk({'row':100000,
                               'chan':128,
                               'x':4096,
                               'y':4096})
        # necessary to make psf optional
        if 'x_psf' in out_ds.dims:
            out_ds = out_ds.chunk({'x_psf': 4096,
                                   'y_psf':4096,
                                   'yo2': 2048})

        dds_out.append(out_ds.unify_chunks())

    writes = xds_to_zarr(dds_out, dds_name, columns='ALL')

    # dask.visualize(writes, color="order", cmap="autumn",
    #                node_attr={"penwidth": "4"},
    #                filename=f'{basename}_grid_ordered_graph.pdf',
    #                optimize_graph=False)
    # dask.visualize(writes, filename=f'{basename}_grid_graph.pdf',
    #                optimize_graph=False)

    print("Computing image space data products", file=log)
    with compute_context(opts.scheduler, basename+'_grid'):
        dask.compute(writes)

    dds = xds_from_zarr(dds_name, chunks={'x': -1, 'y': -1})

    # convert to fits files
    fitsout = []
    if opts.fits_mfs:
        if opts.dirty:
            fitsout.append(dds2fits_mfs(dds, 'DIRTY', basename, norm_wsum=True))
        if opts.psf:
            fitsout.append(dds2fits_mfs(dds, 'PSF', basename, norm_wsum=True))
        if has_model:
            fitsout.append(dds2fits_mfs(dds, 'RESIDUAL', basename, norm_wsum=True))
            fitsout.append(dds2fits_mfs(dds, 'MODEL', basename, norm_wsum=False))

    if opts.fits_cubes:
        if opts.dirty:
            fitsout.append(dds2fits(dds, 'DIRTY', basename, norm_wsum=True))
        if opts.psf:
            fitsout.append(dds2fits(dds, 'PSF', basename, norm_wsum=True))
        if has_model:
            fitsout.append(dds2fits(dds, 'RESIDUAL', basename, norm_wsum=True))
            fitsout.append(dds2fits(dds, 'MODEL', basename, norm_wsum=False))

    if len(fitsout):
        print("Writing fits", file=log)
        dask.compute(fitsout)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)



import numpy as np
import numexpr as ne
from numba import generated_jit, njit, prange
from numba.types import literal
from dask.graph_manipulation import clone
import dask.array as da
from xarray import Dataset
from pfb.operators.gridder import vis2im
from quartical.utils.numba import coerce_literal
from daskms.optimisation import inlined_array
from operator import getitem
from pfb.utils.beam import interp_beam
import dask


def weight_from_sigma(sigma):
    weight = ne.evaluate('1.0/(sigma*sigma)',
                         casting='same_kind')
    return weight


def single_stokes(ds=None,
                  jones=None,
                  opts=None,
                  freq=None,
                  chan_width=None,
                  cell_rad=None,
                  utime=None,
                  tbin_idx=None,
                  tbin_counts=None,
                  radec=None,
                  antpos=None,
                  poltype=None):

    if opts.precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif opts.precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128

    data = getattr(ds, opts.data_column).data
    nrow, nchan, ncorr = data.shape

    ant1 = clone(ds.ANTENNA1.data)
    ant2 = clone(ds.ANTENNA2.data)
    uvw = clone(ds.UVW.data)

    # MS may contain auto-correlations
    if 'FLAG_ROW' in ds:
        frow = clone(ds.FLAG_ROW.data) | (ant1 == ant2)
    else:
        frow = (ant1 == ant2)

    frow = inlined_array(frow, [ant1, ant2])

    if opts.flag_column is not None:
        flag = getattr(ds, opts.flag_column).data
        flag = da.any(flag, axis=2)
        flag = da.logical_or(flag, frow[:, None])
    else:
        flag = da.broadcast_to(frow[:, None], (nrow, nchan))

    if opts.sigma_column is not None:
        sigma = getattr(ds, opts.sigma_column).data
        # weight = 1.0/sigma**2
        weight = da.map_blocks(weight_from_sigma,
                               sigma,
                               chunks=sigma.chunks)
    elif opts.weight_column is not None:
        weight = getattr(ds, opts.weight_column).data
        if opts.weight_column=='WEIGHT':
            weight = da.broadcast_to(weight[:, None, :],
                                     (nrow, nchan, ncorr),
                                     chunks=data.chunks)
    else:
        weight = da.ones_like(data, dtype=real_type)

    if data.dtype != complex_type:
        data = data.astype(complex_type)

    if weight.dtype != real_type:
        weight = weight.astype(real_type)

    if jones is not None:
        if jones.dtype != complex_type:
            jones = jones.astype(complex_type)
        # qcal has chan and ant axes reversed compared to pfb implementation
        jones = da.swapaxes(jones, 1, 2)
    else:
        ntime = utime.size
        nchan = freq.size
        nant = antpos.shape[0]
        jones = da.ones((ntime, nant, nchan, 1, 2),
                        chunks=(-1,)*5,
                        dtype=complex_type)

    # Note we do not chunk at this level since all the chunking happens upfront
    # we cast to dask arrays simply to defer the compute
    tbin_idx = da.from_array(tbin_idx, chunks=(-1))
    tbin_counts = da.from_array(tbin_counts, chunks=(-1))
    vis, wgt = weight_data(data, weight, jones, tbin_idx, tbin_counts,
                        ant1, ant2, pol=poltype, product=opts.product)

    vis = inlined_array(vis, [ant1, ant2, tbin_idx, tbin_counts, jones])
    wgt = inlined_array(wgt, [ant1, ant2, tbin_idx, tbin_counts, jones])

    if isinstance(opts.radec, str):
        raise NotImplementedError()
    elif isinstance(opts.radec, np.ndarray) and not np.array_equal(radec, opts.radec):
        raise NotImplementedError()


    flag = inlined_array(flag, [frow])

    # do this before casting to dask array otherwise
    # serialisation of attrs fails
    freq_out = np.mean(freq)
    freq_min = freq.min()
    freq_max = freq.max()

    # simple average over channels
    if opts.chan_average > 1:
        from africanus.averaging.dask import time_and_channel

        res = time_and_channel(
                    ds.TIME.data,
                    ds.INTERVAL.data,
                    ant1,
                    ant2,
                    uvw=uvw,
                    flag=flag[:, :, None],
                    weight_spectrum=wgt[:, :, None],
                    visibilities=vis[:, :, None],
                    chan_freq=da.from_array(freq, chunks=-1),
                    chan_width=da.from_array(chan_width, chunks=-1),
                    time_bin_secs=1e-15,
                    chan_bin_size=opts.chan_average)

        # map_blocks to deal with nan row chunks.
        # only freq average -> row chunks are preserved
        cchunk = nchan//opts.chan_average
        vis = res.visibilities.map_blocks(lambda x: x,
                                          chunks=(nrow, cchunk, 1))[:, :, 0]
        wgt = res.weight_spectrum.map_blocks(lambda x: x,
                                             chunks=(nrow, cchunk, 1))[:, :, 0]
        flag = res.flag.map_blocks(lambda x: x,
                                   chunks=(nrow, cchunk, 1))[:, :, 0]
        uvw = res.uvw.map_blocks(lambda x: x,
                                 chunks=(nrow, 3))

        freq = res.chan_freq.map_blocks(lambda x: x, chunks=(cchunk,))

    # if opts.bda_decorr < 1:
    #     wgt = da.where(mask, wgt, 0.0)
    #     from africanus.averaging.dask import bda

    #     w_avs = []
    #     uvw = uvw.compute()
    #     t = time.compute()
    #     a1 = ant1.compute()
    #     a2 = ant2.compute()
    #     intv = interval.compute()
    #     fr = frow.compute()[:, None, None]

    #     res = bda(ds.TIME.data,
    #                 ds.INTERVAL.data,
    #                 ant1, ant2,
    #                 uvw=uvw,
    #                 flag=f[:, :, None],
    #                 weight_spectrum=wgt[:, :, None],
    #                 chan_freq=freq,
    #                 chan_width=chan_width,
    #                 decorrelation=0.95,
    #                 min_nchan=freq.size)

    #     uvw = res.uvw.reshape(-1, nchan, 3)[:, 0, :]
    #     wgt = res.weight_spectrum.reshape(-1, nchan).squeeze()

    #     uvw = uvw.rechunk({0:opts.row_out_chunk})

    mask = ~flag

    data_vars = {}
    data_vars['FREQ'] = (('chan',), freq)
    data_vars['WEIGHT'] = (('row', 'chan'), wgt)
    data_vars['UVW'] = (('row', 'uvw'), uvw)
    data_vars['VIS'] = (('row', 'chan'), vis)
    data_vars['MASK'] = (('row', 'chan'), mask.astype(np.uint8))

    # TODO - interpolate beam in time and frequency
    # Instead of BEAM we should have a pre-init step which computes
    # per facet best approximations to smooth beams as described in
    # https://www.overleaf.com/read/yzrsrdwxhxrd
    npix = int(np.deg2rad(opts.max_field_of_view)/cell_rad)
    beam = interp_beam(freq_out/1e6, npix, npix, np.rad2deg(cell_rad), opts.beam_model)
    data_vars['BEAM'] = (('scalar'), beam)

    coords = {'chan': (('chan',), freq)} #,
            #   'row': (('row',), ds.ROWID.values)}

    # TODO - provide time and freq centroids
    attrs = {
        'ra' : radec[0],
        'dec': radec[1],
        'fieldid': ds.FIELD_ID,
        'ddid': ds.DATA_DESC_ID,
        'scanid': ds.SCAN_NUMBER,
        'freq_out': freq_out,
        'freq_min': freq_min,
        'freq_max': freq_max,
        'time_out': np.mean(utime),
        'time_min': utime.min(),
        'time_max': utime.max(),
        'product': opts.product
    }

    out_ds = Dataset(data_vars, coords=coords,
                     attrs=attrs).chunk({'row':100000,
                                         'chan':128})

    return out_ds.unify_chunks()


def weight_data(data, weight, jones, tbin_idx, tbin_counts,
                ant1, ant2, pol='linear', product='I'):
    # data are not necessarily 2x2 so we need separate labels
    # for jones correlations and data/weight correlations
    if jones.ndim == 5:
        jout = 'rafdx'
    elif jones.ndim == 6:
        jout = 'rafdxx'
        # TODO - how do we know if we should return
        # jones[0][0][0] or jones[0][0][0][0] in function wrapper?
        # Not required with delayed
        raise NotImplementedError("Not yet implemented")
    res = da.blockwise(_weight_data, 'rf',
                       data, 'rfc',
                       weight, 'rfc',
                       jones, jout,
                       tbin_idx, 'r',
                       tbin_counts, 'r',
                       ant1, 'r',
                       ant2, 'r',
                       pol, None,
                       product, None,
                       align_arrays=False,
                       meta=np.empty((0, 0), dtype=object))

    vis = da.blockwise(getitem, 'rf', res, 'rf', 0, None, dtype=data.dtype)
    wgt = da.blockwise(getitem, 'rf', res, 'rf', 1, None, dtype=weight.dtype)

    return vis, wgt

def _weight_data(data, weight, jones, tbin_idx, tbin_counts,
                 ant1, ant2, pol, product):
    return _weight_data_impl(data[0], weight[0], jones[0][0][0],
                             tbin_idx, tbin_counts,
                             ant1, ant2, pol, product)

@generated_jit(nopython=True, nogil=True, cache=True, parallel=True)
def _weight_data_impl(data, weight, jones, tbin_idx, tbin_counts,
                      ant1, ant2, pol, product):

    coerce_literal(_weight_data, ["product", "pol"])

    vis_func, wgt_func = stokes_funcs(data, jones, product, pol=pol)

    def _impl(data, weight, jones, tbin_idx, tbin_counts,
              ant1, ant2, pol, product):
        # for dask arrays we need to adjust the chunks to
        # start counting from zero
        tbin_idx -= tbin_idx.min()
        nt = np.shape(tbin_idx)[0]
        nrow, nchan, ncorr = data.shape
        vis = np.zeros((nrow, nchan), dtype=data.dtype)
        wgt = np.zeros((nrow, nchan), dtype=data.real.dtype)

        for t in prange(nt):
            for row in range(tbin_idx[t],
                             tbin_idx[t] + tbin_counts[t]):
                p = int(ant1[row])
                q = int(ant2[row])
                gp = jones[t, p, :, 0]
                gq = jones[t, q, :, 0]
                for chan in range(nchan):
                    wval = wgt_func(gp[chan], gq[chan],
                                    weight[row, chan])
                    if wval > 1e-6:
                        wgt[row, chan] = wval
                        vis[row, chan] = vis_func(gp[chan], gq[chan],
                                                weight[row, chan],
                                                data[row, chan])/wval

        return vis, wgt
    return _impl


def stokes_funcs(data, jones, product, pol):
    # The expressions for DIAG_DIAG and DIAG mode are essentially the same
    if jones.ndim == 5:
        # I and Q have identical weights
        @njit(nogil=True, fastmath=True, inline='always')
        def wfunc(gp, gq, W):
            gp00 = gp[0]
            gp11 = gp[1]
            gq00 = gq[0]
            gq11 = gq[1]
            W0 = W[0]
            W3 = W[-1]
            return np.real(W0*gp00*gq00*np.conjugate(gp00)*np.conjugate(gq00) +
                    W3*gp11*gq11*np.conjugate(gp11)*np.conjugate(gq11))

        if product == literal('I'):
            @njit(nogil=True, fastmath=True, inline='always')
            def vfunc(gp, gq, W, V):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W0 = W[0]
                W3 = W[-1]
                v00 = V[0]
                v11 = V[-1]
                return (W0*gq00*v00*np.conjugate(gp00) +
                        W3*gq11*v11*np.conjugate(gp11))

        elif product == literal('Q'):
            if pol != literal('linear'):
                msg = "Stokes I not currently supported for circular pol"
                raise NotImplementedError(msg)
            @njit(nogil=True, fastmath=True, inline='always')
            def vfunc(gp, gq, W, V):
                gp00 = gp[0]
                gp11 = gp[1]
                gq00 = gq[0]
                gq11 = gq[1]
                W0 = W[0]
                W3 = W[-1]
                v00 = V[0]
                v11 = V[-1]
                return (W0*gq00*v00*np.conjugate(gp00) -
                        W3*gq11*v11*np.conjugate(gp11))

        else:
            raise ValueError("The requested product is not available from input data")

        return vfunc, wfunc

    # Full mode
    elif jones.ndim == 6:
        raise NotImplementedError("Full polarisation imaging not yet supported")

    else:
        raise ValueError("jones array has an unsupported number of dimensions")



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
from functools import partial
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('FWDBWD')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.fwdbwd["inputs"].keys():
    defaults[key] = schema.fwdbwd["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.fwdbwd)
def fwdbwd(**kw):
    '''
    Implements SARA minor cycle using forward backward steps
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'fwdbwd_{timestamp}.log')

    if opts.nworkers is None:
        if opts.scheduler=='distributed':
            opts.nworkers = opts.nband
        else:
            opts.nworkers = 1

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _fwdbwd(**opts)

def _fwdbwd(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    import dask
    import dask.array as da
    import xarray as xr
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    from pfb.utils.fits import load_fits, set_wcs, save_fits, data_from_header
    from pfb.utils.misc import setup_image_data, init_mask
    from pfb.operators.psi import im2coef, coef2im
    from pfb.operators.hessian import hessian, hessian_xds
    from pfb.operators.psf import _hessian_reg_psf
    from pfb.opt.pcg import pcg
    from pfb.opt.primal_dual import primal_dual
    from pfb.opt.power_method import power_method
    from pfb.prox.prox_21m import prox_21m
    from pfb.prox.prox_21 import prox_21
    from pfb.wavelets.wavelets import wavelet_setup
    import pywt
    from copy import deepcopy

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    dds_name = f'{basename}_{opts.postfix}.dds'
    mds_name = f'{basename}_{opts.postfix}.mds'

    dds = xds_from_zarr(dds_name, chunks={'row':opts.row_chunk,
                                          'chan':-1})
    # only a single mds (for now)
    mds = xds_from_zarr(mds_name, chunks={'band':1})[0]
    nband = mds.nband
    nx = mds.nx
    ny = mds.ny
    for ds in dds:
        assert ds.nx == nx
        assert ds.ny == ny

    real_type = dds[0].DIRTY.dtype
    complex_type = np.result_type(real_type, np.complex64)
    if opts.model_name in mds:
        model = mds.get(opts.model_name).values
        assert model.shape == (nband, nx, ny)
        print(f"Initialising model from {opts.model_name} in mds", file=log)
    else:
        print('Initialising model to zeros', file=log)
        model = np.zeros((nband, nx, ny), dtype=real_type)

    mask = init_mask(opts.mask, mds, dds[0].DIRTY.dtype, log)

    # combine images over scan/spw for preconditioner
    dirty = [da.zeros((nx, ny), chunks=(-1, -1),
                         dtype=real_type) for _ in range(nband)]
    wsums = [da.zeros(1, dtype=real_type) for _ in range(nband)]
    nx_psf, ny_psf = dds[0].PSF.shape
    nx_psf, nyo2_psf = dds[0].PSFHAT.shape
    psfhat = [da.zeros((nx_psf, nyo2_psf), chunks=(-1, -1),
                        dtype=complex_type) for _ in range(nband)]
    mean_beam = [da.zeros((nx, ny), chunks=(-1, -1),
                            dtype=real_type) for _ in range(nband)]
    for ds in dds:
        b = ds.bandid
        dirty[b] += ds.DIRTY.data * ds.BEAM.data
        psfhat[b] += ds.PSFHAT.data
        mean_beam[b] += ds.BEAM.data * ds.WSUM.data[0]
        wsums[b] += ds.WSUM.data[0]
    wsums = da.stack(wsums).squeeze()
    wsum = wsums.sum()
    dirty = da.stack(dirty)/wsum
    psfhat = da.stack(psfhat)/wsum
    mean_beam = da.stack(mean_beam)/wsums[:, None, None]
    residual, psfhat, mean_beam, wsum = dask.compute(dirty,
                                                     psfhat,
                                                     mean_beam,
                                                     wsum)

    # dictionary setup
    print("Setting up dictionary", file=log)
    bases = tuple(opts.bases.split(','))
    nbasis = len(bases)
    iys, sys, ntot, nmax = wavelet_setup(residual, bases, opts.nlevels)
    ntot = tuple(ntot)
    psiH = partial(im2coef, bases=bases, ntot=ntot, nmax=nmax,
                   nlevels=opts.nlevels)
    psi = partial(coef2im, bases=bases, ntot=ntot,
                  iy=iys, sy=sys, nx=nx, ny=ny)

    hessopts = {}
    hessopts['cell'] = dds[0].cell_rad
    hessopts['wstack'] = opts.wstack
    hessopts['epsilon'] = opts.epsilon
    hessopts['double_accum'] = opts.double_accum
    hessopts['nthreads'] = opts.nvthreads

    nx_psf, ny_psf = dds[0].nx_psf, dds[0].ny_psf
    npad_xl = (nx_psf - nx)//2
    npad_xr = nx_psf - nx - npad_xl
    npad_yl = (ny_psf - ny)//2
    npad_yr = ny_psf - ny - npad_yl
    psf_padding = ((0,0), (npad_xl, npad_xr), (npad_yl, npad_yr))
    unpad_x = slice(npad_xl, -npad_xr)
    unpad_y = slice(npad_yl, -npad_yr)
    lastsize = ny + np.sum(psf_padding[-1])

    residual_mfs = np.sum(residual, axis=0)
    rms = np.std(residual_mfs)
    rmax = residual_mfs.max()

    lcoord = -(nx//2) + np.arange(nx)
    mcoord = -(ny//2) + np.arange(ny)
    scoord = (lcoord[:, None]**2 + mcoord[None, :]**2).astype(np.float64)
    sigmainv = opts.sigmainv # + nband * (scoord/scoord.max())**4/rms**2
    if np.any(sigmainv):
        M = lambda x: x / sigmainv
    else:
        M = None
    # the PSF is normalised so we don't need to pass wsum
    hess = partial(_hessian_reg_psf, beam=mean_beam * mask[None],
                   psfhat=psfhat, nthreads=opts.nvthreads,
                   sigmainv=sigmainv, padding=psf_padding,
                   unpad_x=unpad_x, unpad_y=unpad_y, lastsize=lastsize)

    if opts.hessnorm is None:
        print("Finding spectral norm of Hessian approximation", file=log)
        hessnorm, hessbeta = power_method(hess, (nband, nx, ny),
                                          tol=opts.pm_tol,
                                          maxit=opts.pm_maxit,
                                          verbosity=opts.pm_verbose,
                                          report_freq=opts.pm_report_freq)
    else:
        hessnorm = opts.hessnorm

    if 'DUAL' in mds:
        dual = mds.DUAL.values
        assert dual.shape == (nband, nbasis, nmax)
    else:
        dual = np.zeros((nband, nbasis, nmax), dtype=model.dtype)

    if 'WEIGHT' in mds:
        weight = mds.WEIGHT.values
        assert weight.shape == (nbasis, nmax)
    else:
        weight = np.ones((nbasis, nmax), dtype=model.dtype)

    # for saving intermediaries
    ra = mds.ra
    dec = mds.dec
    radec = [ra, dec]
    cell_rad = mds.cell_rad
    cell_deg = np.rad2deg(cell_rad)
    freq_out = mds.band.values
    hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq_out)

    print(f"It {0}: max resid = {rmax:.3e}, rms = {rms:.3e}", file=log)
    for i in range(opts.niter):
        print('Getting update', file=log)
        print("Solving for update", file=log)
        if opts.inverter == 'pcg':
            update = pcg(hess_psf,
                        residual,
                        x0=residual/pix_per_beam if k==0 else update,
                        tol=opts.cg_tol,
                        maxit=opts.cg_maxit,
                        minit=opts.cg_minit,
                        verbosity=opts.cg_verbose,
                        report_freq=opts.cg_report_freq,
                        backtrack=opts.backtrack)
        elif k==0:  #opts.inverter == 'pd':
            grad2 = lambda x: hess_psf(x) - residual
            if dual2 is None:
                dual2 = np.zeros_like(residual)
            update, dual2 = primal_dual(residual/pix_per_beam if k==0 else update,
                                        dual2,
                                        rms,
                                        lambda x: x,
                                        lambda x: x,
                                        hessnorm,
                                        prox2,
                                        grad2,
                                        nu=1,
                                        sigma=1,
                                        positivity=dual is None and opts.positivity,  # only at the outset
                                        tol=opts.pd_tol,
                                        maxit=opts.pd_maxit,
                                        verbosity=opts.pd_verbose,
                                        report_freq=opts.pd_report_freq,
                                        gamma=1.0)
        else:
            raise ValueError(f"{opts.inverter} is not a valid inverter")



        # save_fits(basename + f'_update_{k+1}.fits', np.mean(update, axis=0), hdr_mfs)

        save_fits(f'{basename}_update_{i}.fits', update, hdr)
        save_fits(f'{basename}_fwd_resid_{i}.fits', fwd_resid, hdr)

        print('Getting model', file=log)
        modelp = deepcopy(model)
        data = model + update
        model, dual = primal_dual(hess,
                                  data,
                                  model,
                                  dual,
                                  opts.sigma21,
                                  psi,
                                  psiH,
                                  weight,
                                  hessnorm,
                                  prox_21,
                                  nu=nbasis,
                                  positivity=opts.positivity,
                                  tol=opts.pd_tol,
                                  maxit=opts.pd_maxit,
                                  verbosity=opts.pd_verbose,
                                  report_freq=opts.pd_report_freq)

        # reweight
        l2_norm = np.linalg.norm(psiH(model), axis=0)
        for m in range(nbasis):
            weight[m] = opts.alpha/(opts.alpha + l2_norm[m])

        model_dask = da.from_array(model, chunks=(1, -1, -1))

        print('Computing residual', file=log)
        # first write it to disk per dataset
        out_ds = []
        for ds in dds:
            dirty = ds.DIRTY.data
            wgt = ds.WEIGHT.data
            uvw = ds.UVW.data
            freq = ds.FREQ.data
            beam = ds.BEAM.data
            vis_mask = ds.MASK.data
            b = ds.bandid
            # we only want to apply the beam once here
            residual = dirty - hessian(beam * model_dask[b], uvw, wgt,
                                       vis_mask, freq, None, hessopts)
            ds = ds.assign(**{'RESIDUAL': (('x', 'y'), residual)})
            out_ds.append(ds)

        writes = xds_to_zarr(out_ds, dds_name, columns='RESIDUAL')
        dask.compute(writes)

        # reconstruct from disk
        dds = xds_from_zarr(dds_name, chunks={'row':opts.row_chunk})
        residual = [da.zeros((nx, ny), chunks=(-1, -1),
                    dtype=real_type) for _ in range(nband)]
        for ds in dds:
            b = ds.bandid
            # we ave to apply the beam here in case it varies with ds
            # the mask will be applied prior to massing into PCG
            residual[b] += ds.RESIDUAL.data * ds.BEAM.data
        residual = da.stack(residual)/wsum
        residual = residual.compute()
        residual_mfs = np.sum(residual, axis=0)
        rms = np.std(residual_mfs)
        rmax = residual_mfs.max()
        eps = np.linalg.norm(model - modelp) / np.linalg.norm(model)
        print(f"It {i+1}: max resid = {rmax:.3e}, rms = {rms:.3e}, eps = {eps:.3e}",
              file=log)
        if eps < opts.tol:
            break

        # sigmainv = opts.sigmainv #+ (scoord/scoord.max())**4/rms**2
        # M = lambda x: x / sigmainv
        # hessnorm, hessbeta = power_method(hess, (nband, nx, ny),
        #                                   b0=hessbeta,
        #                                   tol=opts.pm_tol,
        #                                   maxit=opts.pm_maxit,
        #                                   verbosity=opts.pm_verbose,
        #                                   report_freq=opts.pm_report_freq)


    print("Saving results", file=log)
    mask = np.any(model, axis=0).astype(bool)
    mask = da.from_array(mask, chunks=(-1, -1))
    model = da.from_array(model, chunks=(1, -1, -1), name=False)
    modelp = da.from_array(modelp, chunks=(1, -1, -1), name=False)
    dual = da.from_array(dual, chunks=(1, -1, -1), name=False)
    weight = da.from_array(weight, chunks=(-1, -1), name=False)

    mds = mds.assign(**{
                     'MASK': (('x', 'y'), mask),
                     'MODEL': (('band', 'x', 'y'), model),
                     'DUAL': (('band', 'basis', 'coef'), dual),
                     'WEIGHT': (('basis', 'coef'), weight)})

    dask.compute(xds_to_zarr(mds, mds_name, columns='ALL'))

    if opts.fits_mfs or opts.fits_cubes:
        print("Writing fits files", file=log)
        dds = xds_from_zarr(dds_name)

        # construct a header from xds attrs
        ra = mds.ra
        dec = mds.dec
        radec = [ra, dec]

        cell_rad = mds.cell_rad
        cell_deg = np.rad2deg(cell_rad)

        freq_out = mds.band.values
        ref_freq = np.mean(freq_out)
        hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)

        model_mfs = np.mean(model, axis=0)
        save_fits(f'{basename}_model_mfs.fits', model_mfs, hdr_mfs)

        residual = np.zeros((nband, nx, ny), dtype=dds[0].DIRTY.dtype)
        wsums = np.zeros(nband)
        for ds in dds:
            b = ds.bandid
            wsums[b] += ds.WSUM.values[0]
            residual[b] += ds.RESIDUAL.values
        wsum = np.sum(wsums)
        residual_mfs = np.sum(residual, axis=0)/wsum
        save_fits(f'{basename}_residual_mfs.fits', residual_mfs, hdr_mfs)

        if opts.fits_cubes:
            # need residual in Jy/beam
            hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, freq_out)
            save_fits(f'{basename}_model.fits', model, hdr)
            fmask = wsums > 0
            residual[fmask] /= wsums[fmask, None, None]
            save_fits(f'{basename}_residual.fits',
                    residual, hdr)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)



import numpy as np
import numexpr as ne
import dask.array as da
from distributed import wait, get_client, as_completed
from operator import getitem
from ducc0.misc import make_noncritical
from pfb.utils.misc import norm_diff
import pyscilog
log = pyscilog.get_logger('PD')


def primal_dual(
        x,  # initial guess for primal variable
        v,  # initial guess for dual variable
        lam,  # regulariser strength
        psi,  # linear operator in dual domain
        psiH,  # adjoint of psi
        L,  # spectral norm of Hessian
        prox,  # prox of regulariser
        grad,  # gradient of smooth term
        nu=1.0,  # spectral norm of psi
        sigma=None,  # step size of dual update
        mask=None,  # regions where mask is False will be masked
        tol=1e-5,
        maxit=1000,
        minit=10,
        positivity=1,
        report_freq=10,
        gamma=1.0,
        verbosity=1):
    # initialise
    xp = x.copy()
    vp = v.copy()
    vtilde = np.zeros_like(v)
    vout = np.zeros_like(v)

    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # start iterations
    eps = 1.0
    k = 0
    while (eps > tol or k < minit) and k < maxit:
        # tmp prox variable
        vtilde = v + sigma * psiH(xp)

        # dual update
        v = vtilde - sigma * prox(vtilde / sigma, lam / sigma)

        # primal update
        x = xp - tau * (psi(2 * v - vp) + grad(xp))
        if positivity == 1:
            x[x < 0.0] = 0.0
        elif positivity == 2:
            msk = np.any(x<=0, axis=0)
            x[:, msk] = 0.0

        # convergence check
        eps = np.linalg.norm(x - xp) / np.linalg.norm(x)

        # copy contents to avoid allocating new memory
        xp[...] = x[...]
        vp[...] = v[...]

        if np.isnan(eps) or np.isinf(eps):
            import pdb; pdb.set_trace()

        if not k % report_freq and verbosity > 1:
            # res = xbar-x
            # phi = np.vdot(res, A(res))
            print(f"At iteration {k} eps = {eps:.3e}",  # and phi = {phi:.3e}",
                  file=log)
        k += 1

    if k == maxit:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations", file=log)

    return x, v


def primal_dual_optimised(
        x,  # initial guess for primal variable
        v,  # initial guess for dual variable
        lam,  # regulariser strength
        psi,  # linear operator in dual domain
        psiH,  # adjoint of psi
        L,  # spectral norm of Hessian
        prox,  # prox of regulariser
        l1weight,
        reweighter,
        grad,  # gradient of smooth term
        nu=1.0,  # spectral norm of psi
        sigma=None,  # step size of dual update
        mask=None,  # regions where mask is False will be masked
        tol=1e-5,
        maxit=1000,
        positivity=1,
        report_freq=10,
        gamma=1.0,
        verbosity=1,
        maxreweight=50):
    # TODO - we can't use make_noncritical because it sometimes returns an
    # array that is not explicitly c-contiguous. How does this impact
    # performance?
    # initialise
    xp = x.copy()
    # xp = make_noncritical(xp)
    vp = v.copy()
    # vp = make_noncritical(vp)
    vtilde = np.zeros_like(v)
    # vtilde = make_noncritical(vtilde)
    vout = np.zeros_like(v)
    # vout = make_noncritical(vout)
    xout = np.zeros_like(x)
    # xout = make_noncritical(xout)


    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # start iterations
    eps = 1.0
    numreweight = 0
    for k in range(maxit):
        # tmp prox variable
        # vtilde = v + sigma * psiH(xp)
        psiH(xp, vtilde)
        ne.evaluate('v + sigma * vtilde', out=vtilde)  #, casting='same_kind')

        # dual update
        # v = vtilde - sigma * prox(vtilde / sigma, lam / sigma)
        prox(vtilde, vout, lam, sigma, weight=l1weight)
        ne.evaluate('vtilde - sigma*vout', out=v)  #, casting='same_kind')

        # primal update
        # x = xp - tau * (psi(2 * v - vp) + grad(xp))
        ne.evaluate('2.0 * v - vp', out=vout)  #, casting='same_kind')
        psi(vout, xout)
        xout += grad(xp)
        ne.evaluate('xp - tau * xout', out=x)  #, casting='same_kind')

        if positivity == 1:
            x[x < 0.0] = 0.0
        elif positivity == 2:
            msk = np.any(x<=0, axis=0)
            x[:, msk] = 0.0

        # convergence check
        eps = np.linalg.norm(x - xp) / np.linalg.norm(x)
        if eps < tol:
            if reweighter is not None and numreweight < maxreweight:
                l1weight = reweighter(x)
                numreweight += 1
            else:
                if numreweight >= maxreweight:
                    print("Maximum reweighting steps reached", file=log)
                break

        # copy contents to avoid allocating new memory
        # this is not faster but it allows us to see where the time is spent
        np.copyto(xp, x)
        np.copyto(vp, v)
        # xp[...] = x[...]
        # vp[...] = v[...]

        if np.isnan(eps) or np.isinf(eps):
            import pdb; pdb.set_trace()

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} eps = {eps:.3e}", file=log)

    if k == maxit-1:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations", file=log)

    return x, v

from pfb.prox.prox_21m import dual_update_numba
def primal_dual_optimised2(
        x,  # initial guess for primal variable
        v,  # initial guess for dual variable
        lam,  # regulariser strength
        psi,  # linear operator in dual domain
        psiH,  # adjoint of psi
        L,  # spectral norm of Hessian
        prox,  # prox of regulariser
        l1weight,
        reweighter,
        grad,  # gradient of smooth term
        nu=1.0,  # spectral norm of psi
        sigma=None,  # step size of dual update
        mask=None,  # regions where mask is False will be masked
        tol=1e-5,
        maxit=1000,
        positivity=1,
        report_freq=10,
        gamma=1.0,
        verbosity=1,
        maxreweight=50):
    # TODO - we can't use make_noncritical because it sometimes returns an
    # array that is not explicitly c-contiguous. How does this impact
    # performance?
    # initialise
    xp = x.copy()
    # xp = make_noncritical(xp)
    vp = v.copy()
    xout = np.zeros_like(x)
    # xout = make_noncritical(xout)


    # this seems to give a good trade-off between
    # primal and dual problems
    if sigma is None:
        sigma = L / (2.0 * gamma) / nu

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # start iterations
    eps = 1.0
    numreweight = 0
    for k in range(maxit):
        psiH(xp, v)
        dual_update_numba(vp, v, lam, sigma=sigma, weight=l1weight)
        ne.evaluate('2.0 * v - vp', out=vp)  #, casting='same_kind')
        psi(vp, xout)
        xout += grad(xp)
        ne.evaluate('xp - tau * xout', out=x)  #, casting='same_kind')

        if positivity == 1:
            x[x < 0.0] = 0.0
        elif positivity == 2:
            msk = np.any(x<=0, axis=0)
            x[:, msk] = 0.0

        # convergence check
        # LB - why is numpy timing so inconsistent?
        # eps = np.linalg.norm(x - xp) / np.linalg.norm(x)
        eps = norm_diff(x, xp)
        if eps < tol:
            if reweighter is not None and numreweight < maxreweight:
                l1weight = reweighter(x)
                numreweight += 1
            else:
                if numreweight >= maxreweight:
                    print("Maximum reweighting steps reached", file=log)
                break

        # copy contents to avoid allocating new memory
        # this is not faster but it allows us to see where the time is spent
        np.copyto(xp, x)
        np.copyto(vp, v)
        # xp[...] = x[...]
        # vp[...] = v[...]

        if np.isnan(eps) or np.isinf(eps):
            import pdb; pdb.set_trace()

        if not k % report_freq and verbosity > 1:
            print(f"At iteration {k} eps = {eps:.3e}", file=log)

    if k == maxit-1:
        if verbosity:
            print(f"Max iters reached. eps = {eps:.3e}", file=log)
    else:
        if verbosity:
            print(f"Success, converged after {k} iterations", file=log)

    return x, v


def vtilde_update(A, sigma, psiH):
    return A.dual + sigma * psiH(A.model)


def get_ratio(vtildes, lam, sigma, l1weights):
    vmfs = np.zeros(l1weights.shape)
    nband = 0
    for wid, v in vtildes.items():
        # l2_norm += (v/sigma)**2
        vmfs += v/sigma
    # l2_norm = np.sqrt(l2_norm)
    vsoft = np.maximum(np.abs(vmfs) - lam*l1weights/sigma, 0.0)  # norm is positive
    # l2_soft = np.where(np.abs(l2_norm) >= lam/sigma, l2_norm, 0.0)
    mask = vmfs != 0
    ratio = np.zeros(mask.shape, dtype=l1weights.dtype)
    ratio[mask] = vsoft[mask] / vmfs[mask]
    return ratio

def update(A, y, vtilde, ratio, xp, vp, sigma, lam, tau, gamma, psi, psiH, positivity):
    # dual
    v = vtilde - vtilde * ratio

    # primal
    grad = -A(y - xp)/gamma
    x = xp - tau * (psi(2 * v - vp) + grad)
    if positivity:
        x[x < 0] = 0.0

    vtilde = v + sigma * psiH(x)

    eps = np.linalg.norm(x-xp)/np.linalg.norm(x)

    return x, v, vtilde, eps


def sety(A, x, gamma):
    if hasattr(A, 'model'):
        return A.model + gamma * x
    else:
        return gamma * x


def primal_dual_dist(
            Afs,
            xfs,
            psi,
            psiH,
            lam,  # regulariser strength,
            L,  # spectral norm of Hessian
            l1weight,
            nu=1.0,  # spectral norm of dictionary
            sigma=None,  # step size of dual update
            tol=1e-5,
            maxit=100,
            positivity=True,
            gamma=1.0,
            verbosity=1):

    client = get_client()

    names = [w['name'] for w in client.scheduler_info()['workers'].values()]

    if sigma is None:
        sigma = L / (2.0 * gamma)

    # stepsize control
    tau = 0.9 / (L / (2.0 * gamma) + sigma * nu**2)

    # we need to do this upfront only at the outset
    yfs = {}
    vtildefs = {}
    modelfs = {}
    dualfs = {}
    for wid, A in Afs.items():
        yf = client.submit(sety, A, xfs[wid], gamma, workers={wid})
        yfs[wid] = yf
        vtildef = client.submit(vtilde_update, A, sigma, psiH, workers={wid})
        vtildefs[wid] = vtildef
        modelf = client.submit(getattr, A, 'model', workers={wid})
        modelfs[wid] = modelf
        dualf = client.submit(getattr, A, 'dual', workers={wid})
        dualfs[wid] = dualf

    epsfs = {}
    for k in range(maxit):
        # TODO - should get one ratio per basis
        # split over different workers
        ratio = client.submit(get_ratio,
                              vtildefs, lam, sigma, l1weight,
                              workers={names[0]})

        wait([ratio])
        for wid, A in Afs.items():
            future = client.submit(update,
                                   A, yfs[wid], vtildefs[wid], ratio,
                                   modelfs[wid], dualfs[wid],
                                   sigma,
                                   lam,
                                   tau,
                                   gamma,
                                   psi,
                                   psiH,
                                   positivity,
                                   workers={wid})
            modelfs[wid] = client.submit(getitem, future, 0, workers={wid})
            dualfs[wid] = client.submit(getitem, future, 1, workers={wid})
            vtildefs[wid] = client.submit(getitem, future, 2, workers={wid})
            epsfs[wid] = client.submit(getitem, future, 3, workers={wid})

        wait(list(epsfs.values()))

        eps = []
        for wid, epsf in epsfs.items():
            eps.append(epsf.result())
        eps = np.array(eps)
        if eps.max() < tol:
            break

    if k >= maxit-1:
        print(f'Maximum iterations reached. eps={eps}', file=log)
    else:
        print(f'Success after {k} iterations', file=log)

    return modelfs, dualfs


primal_dual.__doc__ = r"""
    Algorithm to solve problems of the form

    argmin_x (xbar - x).T A (xbar - x)/2 + lam |PSI.H x|_21 s.t. x >= 0

    where x is the image cube and PSI an orthogonal basis for the spatial axes
    and the positivity constraint is optional.

    A        - Positive definite Hermitian operator
    xbar     - Data-like term
    x0       - Initial guess for primal variable
    v0       - Initial guess for dual variable
    lam      - Strength of l21 regulariser
    psi      - Orthogonal basis operator where
               psi.hdot() does decomposition and
               psi.dot() the reconstruction
               for all bases and channels
    weights  - Weights for l1 thresholding
    L        - Lipschitz constant of A
    nu       - Spectral norm of psi
    sigma    - l21 step size (set to L/2 by default)
    gamma    - step size during forward step

    Note that the primal variable (i.e. x) has shape (nband, nx, ny) where
    nband is the number of imaging bands and the dual variable (i.e. v) has
    shape (nbasis, nband, nmax) where nmax is the total number of coefficients
    for all bases. It is assumed that all bases are decomposed into the same
    number of coefficients. We use zero padding to deal with the fact that the
    basess do not necessarily have the same number of coefficients.
    """



import numpy as np
import dask.array as da
from pfb.utils.misc import fitcleanbeam

# submit on these
def get_resid_and_stats(Afs, wsum):
    dirty = list(Afs.values())[0].dirty
    nx, ny = dirty.shape
    residual_mfs = np.zeros((nx, ny), dtype=dirty.dtype)
    for wid, A in Afs.items():
        if hasattr(A, 'residual'):
            residual_mfs += A.residual
        else:
            residual_mfs += A.dirty
    residual_mfs /= wsum
    rms = np.std(residual_mfs)
    rmax = np.sqrt((residual_mfs**2).max())
    return residual_mfs, rms, rmax

def accum_wsums(As):
    wsum = 0
    for wid, A in As.items():
        wsum += A.wsumb
    return wsum

def get_epsb(xp, x):
    return np.sum((x-xp)**2), np.sum(x**2)

def get_eps(num, den):
    return np.sqrt(np.sum(num)/np.sum(den))

def l1reweight(model, residual, l1weight, psiH, wsum, pix_per_beam, alpha=2):
    # get norm of model and residual
    l2mod = np.zeros(l1weight.shape, l1weight.dtype)
    l2res = np.zeros(l1weight.shape, l1weight.dtype)
    nbasis = l1weight.shape[0]
    nband = 0
    for wid, mod in model.items():
        l2mod += psiH(mod)
        # Jy/beam -> Jy/pixel
        l2res += psiH(residual[wid]/wsum/pix_per_beam)
        nband += 1
    l2mod /= nband
    rms = np.std(l2res, axis=-1)
    return alpha/(1 + (l2mod/rms[:, None])**2)


def set_l1weight(ds):
    return ds.L1WEIGHT.values


def get_cbeam_area(As, wsum):
    psf_mfs = np.zeros(As[0].psf.shape, dtype=As[0].psf.dtype)
    for wid, A in As.items():
        psf_mfs += A.psf
    psf_mfs /= wsum
    # beam pars in pixel units
    GaussPar = fitcleanbeam(psf_mfs[None], level=0.5, pixsize=1.0)[0]
    return GaussPar[0]*GaussPar[1]*np.pi/4


def update_results(A, ds, model, dual, residual):
    A.model[...] = model
    A.dual[...] = dual
    A.residual[...] = residual

    ds_out = ds.assign(**{'MODEL': (('x','y'), da.from_array(model, chunks=(4096, 4096))),
                          'DUAL': (('b', 'c'), da.from_array(dual, chunks=(1, 4096**2))),
                          'RESIDUAL': (('x', 'y'), da.from_array(residual, chunks=(4096, 4096)))})
    return ds_out


from pfb.operators.hessian import _hessian_impl
def compute_residual(A, model):
    return A.compute_residual(model)


def set_wsum(A, wsum):
    A.wsum = wsum
    return



import numpy as np
from numba import njit, prange
import dask.array as da
from ducc0.fft import c2c
from africanus.constants import c as lightspeed
from quartical.utils.dask import Blocker
iFs = np.fft.ifftshift
Fs = np.fft.fftshift


def compute_counts(uvw, freq, mask, nx, ny,
                   cell_size_x, cell_size_y, dtype, k=6, ngrid=1):

    counts = da.blockwise(compute_counts_wrapper, ('row', 'nx', 'ny'),
                          uvw, ('row', 'three'),
                          freq, ('chan',),
                          mask, ('row', 'chan'),
                          nx, None,
                          ny, None,
                          cell_size_x, None,
                          cell_size_y, None,
                          dtype, None,
                          k, None,
                          ngrid, None,
                          new_axes={"nx": nx, "ny": ny},
                          adjust_chunks={'row': ngrid},
                          align_arrays=False,
                          dtype=dtype)

    return counts.sum(axis=0)


def compute_counts_wrapper(uvw, freq, mask, nx, ny,
                           cell_size_x, cell_size_y, dtype, k, ngrid):
    return _compute_counts(uvw[0], freq[0], mask[0], nx, ny,
                        cell_size_x, cell_size_y, dtype, k, ngrid)



@njit(nogil=True, fastmath=True, cache=True, parallel=True)
def _compute_counts(uvw, freq, mask, nx, ny,
                    cell_size_x, cell_size_y, dtype,
                    k=6, ngrid=1):  # support hardcoded for now
    # ufreq
    u_cell = 1/(nx*cell_size_x)
    # shifts fftfreq such that they start at zero
    # convenient to look up the pixel value
    umax = np.abs(-1/cell_size_x/2 - u_cell/2)

    # vfreq
    v_cell = 1/(ny*cell_size_y)
    vmax = np.abs(-1/cell_size_y/2 - v_cell/2)

    # initialise array to store counts
    # the additional axis is to allow chunking over row
    counts = np.zeros((ngrid, nx, ny), dtype=dtype)

    # accumulate counts
    nrow = uvw.shape[0]
    nchan = freq.size
    bin_counts = [nrow // ngrid + (1 if x < nrow % ngrid else 0)  for x in range (ngrid)]
    bin_idx = np.zeros(ngrid, dtype=np.int64)
    bin_counts = np.asarray(bin_counts).astype(bin_idx.dtype)
    bin_idx[1:] = np.cumsum(bin_counts)[0:-1]

    normfreq = freq / lightspeed
    ko2 = k//2
    ko2sq = ko2**2

    for g in prange(ngrid):
        for r in range(bin_idx[g], bin_idx[g] + bin_counts[g]):
            uvw_row = uvw[r]
            for c in range(nchan):
                if not mask[r, c]:
                    continue
                # current uv coords
                chan_normfreq = normfreq[c]
                u_tmp = uvw_row[0] * chan_normfreq
                v_tmp = uvw_row[1] * chan_normfreq
                # pixel coordinates
                ug = (u_tmp + umax)/u_cell
                vg = (v_tmp + vmax)/v_cell
                if k:
                    # indices
                    u_idx = int(np.round(ug))
                    v_idx = int(np.round(vg))
                    for i in range(-ko2, ko2):
                        x_idx = i + u_idx
                        x = x_idx - ug + 0.5
                        val = _es_kernel(x/ko2, 2.3, k)
                        for j in range(-ko2, ko2):
                            y_idx = j + v_idx
                            y = y_idx - vg + 0.5
                            counts[g, x_idx, y_idx] += val * _es_kernel(y/ko2, 2.3, k)
                else:  # nearest neighbour
                    # indices
                    u_idx = int(np.floor(ug))
                    v_idx = int(np.floor(vg))
                    counts[g, u_idx, v_idx] += 1.0
    return counts  #.sum(axis=0, keepdims=True)

@njit(nogil=True, fastmath=True, cache=True, inline='always')
def _es_kernel(x, beta, k):
    return np.exp(beta*k*(np.sqrt((1-x)*(1+x)) - 1))

def counts_to_weights(counts, uvw, freq, nx, ny,
                      cell_size_x, cell_size_y, robust):

    weights = da.blockwise(counts_to_weights_wrapper, ('row', 'chan'),
                           counts, ('nx', 'ny'),
                           uvw, ('row', 'three'),
                           freq, ('chan',),
                           nx, None,
                           ny, None,
                           cell_size_x, None,
                           cell_size_y, None,
                           robust, None,
                           dtype=counts.dtype)
    return weights

def counts_to_weights_wrapper(counts, uvw, freq, nx, ny,
                              cell_size_x, cell_size_y, robust):
    return _counts_to_weights(counts[0][0], uvw[0], freq, nx, ny,
                              cell_size_x, cell_size_y, robust)


@njit(nogil=True, fastmath=True, cache=True)
def _counts_to_weights(counts, uvw, freq, nx, ny,
                       cell_size_x, cell_size_y, robust):
    # ufreq
    u_cell = 1/(nx*cell_size_x)
    umax = np.abs(-1/cell_size_x/2 - u_cell/2)

    # vfreq
    v_cell = 1/(ny*cell_size_y)
    vmax = np.abs(-1/cell_size_y/2 - v_cell/2)

    # initialise array to store counts
    # the additional axis is to allow chunking over row
    nchan = freq.size
    nrow = uvw.shape[0]

    weights = np.zeros((nrow, nchan), dtype=counts.dtype)
    if not counts.any():
        return weights

    # Briggs weighting factor
    if robust > -2:
        numsqrt = 5*10**(-robust)
        avgW = (counts ** 2).sum() / counts.sum()
        ssq = numsqrt * numsqrt/avgW
        counts = 1 + counts * ssq

    normfreq = freq / lightspeed
    for r in range(nrow):
        uvw_row = uvw[r]
        for c in range(nchan):
            # get current uv
            chan_normfreq = normfreq[c]
            u_tmp = uvw_row[0] * chan_normfreq
            v_tmp = uvw_row[1] * chan_normfreq
            # get u index
            u_idx = int(np.floor((u_tmp + umax)/u_cell))
            # get v index
            v_idx = int(np.floor((v_tmp + vmax)/v_cell))
            if counts[u_idx, v_idx]:
                weights[r, c] = 1.0/counts[u_idx, v_idx]
    return weights


def filter_extreme_counts(counts, nbox=16, nlevel=10):

    return da.blockwise(_filter_extreme_counts, 'xy',
                        counts, 'xy',
                        nbox, None,
                        nlevel, None,
                        dtype=counts.dtype)



@njit(nogil=True, fastmath=True, cache=True)
def _filter_extreme_counts(counts, nbox=16, level=10):
    '''
    Replaces extreme counts by local mean computed i
    '''
    nx, ny = counts.shape
    I, J = np.where(counts>0)
    for i, j in zip(I, J):
        ilow = np.maximum(0, i-nbox//2)
        ihigh = np.minimum(nx, i+nbox//2)
        jlow = np.maximum(0, j-nbox//2)
        jhigh = np.minimum(ny, j+nbox//2)
        tmp = counts[ilow:ihigh, jlow:jhigh]
        ix, iy = np.where(tmp)
        # check if there are too few values to compare to
        if ix.size < nbox:
            counts[i, j] = 0
            continue
        local_mean = np.mean(tmp[ix, iy])
        if counts[i,j] < local_mean/level:
            counts[i, j] = local_mean
    return counts


# from scipy.special import polygamma
# import dask
# @dask.delayed()
# def etaf(ressq, ovar, dof):
#     eta = (dof + 1)/(dof + ressq/ovar)
#     logeta = polygamma(0, dof+1) - np.log(dof + ressq/ovar)
#     return eta, logeta


# from pfb.operators.gridder import im2vis
# def l2reweight(dsv, dsi, epsilon, nthreads, wstack, precision, dof=2):
#     # vis data products
#     uvw = dsv.UVW.data
#     freq = dsv.FREQ.data
#     vis = dsv.VIS.data
#     vis_mask = dsv.MASK.data

#     # image data products
#     model = dsi.MODEL.data
#     cell_rad = dsi.cell_rad
#     x0 = dsi.x0
#     y0 = dsi.y0
#     beam = dsi.BEAM.data

#     # residual from model visibilities
#     mvis = im2vis(uvw=uvw,
#                     freq=freq,
#                     image=model*beam,
#                     cellx=cell_rad,
#                     celly=cell_rad,
#                     nthreads=nthreads,
#                     epsilon=epsilon,
#                     do_wgridding=wstack,
#                     x0=x0,
#                     y0=y0,
#                     precision=precision)
#     res = vis - mvis
#     res *= vis_mask

#     # Mahalanobis distance
#     ressq = (res*res.conj()).real

#     # overall variance factor
#     eta = da.map_blocks(update_eta,
#                         ressq,
#                         vis_mask,
#                         dof,
#                         chunks=ressq.chunks)

#     return eta, res


# def update_eta(ressq, vis_mask, dof):
#     wcount = vis_mask.sum()
#     if wcount:
#         ovar = ressq.sum()/wcount
#         eta = (dof + 1)/(dof + ressq/ovar)
#         return eta/ovar
#     else:
#         return np.zeros_like(ressq)



import numpy as np
from africanus.model.spi import fit_spi_components
from numpy.testing._private.utils import assert_allclose
from pfb.utils.misc import convolve2gaussres, Gaussian2D
import pytest

pmp = pytest.mark.parametrize

@pmp("nx", [128])
@pmp("ny", [80, 220])
@pmp("nband", [4, 8])
@pmp("alpha", [-0.5, 0.0, 0.5])
def test_convolve2gaussres(nx, ny, nband, alpha):
    np.random.seed(420)
    freq = np.linspace(0.5e9, 1.5e9, nband)
    ref_freq = freq[0]

    Gausspari = ()
    es = np.linspace(15, 5, nband)
    for v in range(nband):
        Gausspari += ((es[v], es[v], 0.0),)

    x = np.arange(-nx/2, nx/2)
    y = np.arange(-ny/2, ny/2)
    xx, yy = np.meshgrid(x, y, indexing='ij')
    restored = np.zeros((nband, nx, ny))
    for v in range(nband):
        restored[v] = Gaussian2D(xx, yy, Gausspari[v],
                                 normalise=False) * (freq[v]/ref_freq)**alpha

    conv_model = convolve2gaussres(restored, xx, yy, Gausspari[0], 8,
                                   gausspari=Gausspari)

    Ix, Iy = np.where(conv_model[-1] > 0.05)

    comps = conv_model[:, Ix, Iy]
    weights = np.ones((nband))

    out = fit_spi_components(comps.T, weights, freq, ref_freq,
                             tol=1e-7, maxiter=250)

    # offset for relative difference
    assert_allclose(1+alpha, 1+out[0, :], atol=5e-4, rtol=5e-4)
    assert_allclose(out[2, :], restored[0, Ix, Iy], atol=5e-4, rtol=5e-4)



import numpy as np
from typing import Tuple
from pfb.utils.misc import Gaussian2D, convolve2gaussres


def restore_image(model: np.ndarray,
                  residual: np.ndarray,
                  cell_size_x: float,
                  cell_size_y: float,
                  gaussparf: Tuple[Tuple[float]],
                  gausspari: Tuple[Tuple[float]],
                  convolve_residuals: bool,
                  nthreads: int,
                  padding_frac: float):
    '''
    Create a restored image with resolution guassparf.
    Optional - convolve residuals to common resolution before adding them
    back in.

    model - model image cube
    residual - residual image cube
    cell_size_x - cell size along l dimension in degrees
    cell_size_y - cell size along m dimension in degrees
    gaussparf - tuple containing Gaussian parameters specifying
        desired resolution
    gaussparf - tuple containing Gaussian parameters specifying
        resolution of residual
    convolve_residuals - whether to also convolve the residuals
        to a common resolution
    '''

    try:
        assert model.ndim == 3
        assert model.shape == residual.shape
        assert len(gaussparf) == model.shape[0]
        assert len(gausspari) == model.shape[0]
    except Exception as e:
        raise e

    nband, nx, ny = model.shape
    x = np.arange(-(nx//2), nx//2 + nx % 2) * cell_size_x
    y = np.arange(-(ny//2), ny//2 + ny % 2) * cell_size_y
    xx, yy = np.meshgrid(x, y)

    for b in range(nband):
        model[b:b+1] = convolve2gaussres(model[b:b+1], xx, yy,
                                         gaussparf[b], nthreads,
                                         norm_kernel=False,  # peak of kernel set to unity
                                         pfrac=padding_frac)

    if convolve_residuals:
        residual = convolve2gaussres(residual, xx, yy, gaussparf[0], nthreads,
                                     gausspari=gausspari,
                                     norm_kernel=True,  # kernel has unit volume
                                     pfrac=padding_frac)

    return model + residual





# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('INIT')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.init["inputs"].keys():
    defaults[key] = schema.init["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.init)
def init(**kw):
    '''
    Initialise data products for imaging
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'init_{timestamp}.log')
    from daskms.fsspec_store import DaskMSStore
    msstore = DaskMSStore(opts.ms.rstrip('/'))
    ms = msstore.fs.glob(opts.ms.rstrip('/'))
    try:
        assert len(ms) > 0
        opts.ms = list(map(msstore.fs.unstrip_protocol, ms))
    except:
        raise ValueError(f"No MS at {opts.ms}")

    if opts.gain_table is not None:
        gainstore = DaskMSStore(opts.gain_table.rstrip('/'))
        gt = gainstore.fs.glob(opts.gain_table.rstrip('/'))
        try:
            assert len(gt) > 0
            opts.gain_table = list(map(gainstore.fs.unstrip_protocol, gt))
        except Exception as e:
            raise ValueError(f"No gain table  at {opts.gain_table}")

    if opts.product.upper() not in ["I"]:
                                    # , "Q", "U", "V", "XX", "YX", "XY",
                                    # "YY", "RR", "RL", "LR", "LL"]:
        raise NotImplementedError(f"Product {opts.product} not yet supported")

    OmegaConf.set_struct(opts, True)

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _init(**opts)

def _init(**kw):
    opts = OmegaConf.create(kw)
    from omegaconf import ListConfig
    if (not isinstance(opts.ms, list) and not
        isinstance(opts.ms, ListConfig)):
        opts.ms = [opts.ms]
    if opts.gain_table is not None:
        if (not isinstance(opts.gain_table, list) and not
            isinstance(opts.gain_table,ListConfig)):
            opts.gain_table = [opts.gain_table]
    OmegaConf.set_struct(opts, True)

    import os
    from pathlib import Path
    import numpy as np
    from pfb.utils.misc import construct_mappings
    import dask
    dask.config.set(**{'array.slicing.split_large_chunks': False})
    from daskms import xds_from_storage_ms as xds_from_ms
    from daskms import xds_from_storage_table as xds_from_table
    from daskms.experimental.zarr import xds_to_zarr, xds_from_zarr
    from daskms.fsspec_store import DaskMSStore
    import dask.array as da
    from africanus.constants import c as lightspeed
    from ducc0.fft import good_size
    from pfb.utils.stokes import single_stokes
    from pfb.utils.correlations import single_corr
    from pfb.utils.misc import compute_context, chunkify_rows
    import xarray as xr

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    xdsstore = DaskMSStore(f'{basename}.xds')
    if xdsstore.exists():
        if opts.overwrite:
            print(f"Overwriting {basename}.xds", file=log)
            xdsstore.rm(recursive=True)
        else:
            raise ValueError(f"{basename}.xds exists. "
                             "Set overwrite to overwrite it. ")

    if opts.gain_table is not None:
        tmpf = lambda x: x.rstrip('/') + f'::{opts.gain_term}'
        gain_names = list(map(tmpf, opts.gain_table))
    else:
        gain_names = None

    if opts.freq_range is not None:
        fmin, fmax = opts.freq_range.strip(' ').split(':')
        if len(fmin) > 0:
            freq_min = float(fmin)
        else:
            freq_min = -np.inf
        if len(fmax) > 0:
            freq_max = float(fmax)
        else:
            freq_max = np.inf
    else:
        freq_min = -np.inf
        freq_max = np.inf

    print('Constructing mapping', file=log)
    row_mapping, freq_mapping, time_mapping, \
        freqs, utimes, ms_chunks, gain_chunks, radecs, \
        chan_widths, uv_max, antpos, poltype = \
            construct_mappings(opts.ms,
                               gain_names,
                               ipi=opts.integrations_per_image,
                               cpi=opts.channels_per_image,
                               freq_min=freq_min,
                               freq_max=freq_max)

    max_freq = 0
    for ms in opts.ms:
        for idt in freqs[ms].keys():
            freq = freqs[ms][idt]
            max_freq = np.maximum(max_freq, freq.max())

    # cell size
    cell_rad = 1.0 / (2 * uv_max * max_freq / lightspeed)

    # # we should rephase to the Barycenter of all datasets
    # if opts.radec is not None:
    #     raise NotImplementedError()

    # this is not optional, concatenate during gridding stage if desired
    group_by = ['FIELD_ID', 'DATA_DESC_ID', 'SCAN_NUMBER']

    # assumes measurement sets have the same columns
    columns = (opts.data_column,
               opts.flag_column,
               'UVW', 'ANTENNA1',
               'ANTENNA2', 'TIME', 'INTERVAL', 'FLAG_ROW')
    schema = {}
    schema[opts.data_column] = {'dims': ('chan', 'corr')}
    schema[opts.flag_column] = {'dims': ('chan', 'corr')}

    # only WEIGHT column gets special treatment
    # any other column must have channel axis
    if opts.sigma_column is not None:
        print(f"Initialising weights from {opts.sigma_column} column", file=log)
        columns += (opts.sigma_column,)
        schema[opts.sigma_column] = {'dims': ('chan', 'corr')}
    elif opts.weight_column is not None:
        print(f"Using weights from {opts.weight_column} column", file=log)
        columns += (opts.weight_column,)
        # hack for https://github.com/ratt-ru/dask-ms/issues/268
        if opts.weight_column != 'WEIGHT':
            schema[opts.weight_column] = {'dims': ('chan', 'corr')}
    else:
        print(f"No weights provided, using unity weights", file=log)

    out_datasets = []
    row_id_start = 0
    for ims, ms in enumerate(opts.ms):
        xds = xds_from_ms(ms, chunks=ms_chunks[ms], columns=columns,
                          table_schema=schema, group_cols=group_by)

        if opts.gain_table is not None:
            gds = xds_from_zarr(gain_names[ims],
                                chunks=gain_chunks[ms])

        for ids, ds in enumerate(xds):
            fid = ds.FIELD_ID
            ddid = ds.DATA_DESC_ID
            scanid = ds.SCAN_NUMBER
            # TODO - cleaner syntax
            if opts.fields is not None:
                fields = opts.fields.strip(' ')  # strip white space
                if fid not in list(map(int, fields[1:-1].split(','))):
                    continue
            if opts.ddids is not None:
                ddids = opts.ddids.strip(' ')
                if ddid not in list(map(int, ddids[1:-1].split(','))):
                    continue
            if opts.scans is not None:
                scans = opts.scans.strip(' ')
                if scanid not in list(map(int, scans[1:-1].split(','))):
                    continue


            idt = f"FIELD{fid}_DDID{ddid}_SCAN{scanid}"
            nrow = ds.dims['row']
            ncorr = ds.dims['corr']

            idx = (freqs[ms][idt]>=freq_min) & (freqs[ms][idt]<=freq_max)
            if not idx.any():
                continue

            nchan = idx.sum()

            for ti, (tlow, tcounts) in enumerate(zip(time_mapping[ms][idt]['start_indices'],
                                           time_mapping[ms][idt]['counts'])):

                It = slice(tlow, tlow + tcounts)
                ridx = row_mapping[ms][idt]['start_indices'][It]
                rcnts = row_mapping[ms][idt]['counts'][It]
                # select all rows for output dataset
                Irow = slice(ridx[0], ridx[-1] + rcnts[-1])

                for flow, fcounts in zip(freq_mapping[ms][idt]['start_indices'],
                                         freq_mapping[ms][idt]['counts']):
                    Inu = slice(flow, flow + fcounts)

                    subds = ds[{'row': Irow, 'chan': Inu}]
                    if opts.gain_table is not None:
                        # Only DI gains currently supported
                        jones = gds[ids][{'gain_time': It, 'gain_freq': Inu}].gains.data
                    else:
                        jones = None

                    if opts.product.upper() in ["I", "Q", "U", "V"]:
                        out_ds = single_stokes(
                            ds=subds,
                            jones=jones,
                            opts=opts,
                            freq=freqs[ms][idt][Inu],
                            chan_width=chan_widths[ms][idt][Inu],
                            utime=utimes[ms][idt][It],
                            tbin_idx=ridx,
                            tbin_counts=rcnts,
                            cell_rad=cell_rad,
                            radec=radecs[ms][idt],
                            antpos=antpos[ms],
                            poltype=poltype[ms])
                    elif opts.product.upper() in ["XX", "YX", "XY", "YY",
                                                  "RR", "RL", "LR", "LL"]:
                        out_ds = single_corr(
                            ds=subds,
                            jones=jones,
                            opts=opts,
                            freq=freqs[ms][idt][Inu],
                            chan_width=chan_widths[ms][idt][Inu],
                            utimes=utimes[ms][idt][It],
                            tbin_idx=ridx,
                            tbin_counts=rcnts,
                            cell_rad=cell_rad,
                            radec=radecs[ms][idt])
                    else:
                        raise NotImplementedError(f"Product {args.product} not "
                                                "supported yet")
                    # if all data in a dataset is flagged we return None and
                    # ignore this chunk of data
                    if out_ds is not None:
                        out_datasets.append(out_ds)

    if len(out_datasets):
        writes = xds_to_zarr(out_datasets, f'{basename}.xds',
                             columns='ALL')
    else:
        raise ValueError('No datasets found to write. '
                         'Data completely flagged maybe?')

    # dask.visualize(writes, color="order", cmap="autumn",
    #                node_attr={"penwidth": "4"},
    #                filename=basename + '_writes_I_ordered_graph.pdf',
    #                optimize_graph=False)
    # dask.visualize(writes, filename=basename +
    #                '_writes_I_graph.pdf', optimize_graph=False)

    with compute_context(opts.scheduler, basename+'_init'):
        dask.compute(writes, optimize_graph=False)

    if opts.scheduler=='distributed':
        from distributed import get_client
        client = get_client()
        client.close()

    print("All done here.", file=log)



# flake8: noqa
import os
from contextlib import ExitStack
from pfb.workers.main import cli
from functools import partial
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('SPOTLESS')

from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.spotless["inputs"].keys():
    defaults[key] = schema.spotless["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.spotless)
def spotless(**kw):
    '''
    Spotless algorithm
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    OmegaConf.set_struct(opts, True)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'spotless_{timestamp}.log')

    with ExitStack() as stack:
        # numpy imports have to happen after this step
        from pfb import set_client
        set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        if opts.scheduler=='distributed':
            return _spotless_dist(**opts)
        else:
            return _spotless(**opts)


def _spotless(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import numpy as np
    import xarray as xr
    import numexpr as ne
    import dask
    import dask.array as da
    from dask.distributed import performance_report
    from pfb.utils.fits import (set_wcs, save_fits, dds2fits,
                                dds2fits_mfs, load_fits)
    from pfb.utils.misc import dds2cubes
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    from pfb.opt.power_method import power_method
    from pfb.opt.pcg import pcg
    from pfb.opt.primal_dual import primal_dual_optimised2 as primal_dual
    from pfb.utils.misc import l1reweight_func
    from pfb.operators.hessian import hessian_xds
    from pfb.operators.psf import psf_convolve_cube
    from pfb.operators.psi import im2coef
    from pfb.operators.psi import coef2im
    from copy import copy, deepcopy
    from ducc0.misc import make_noncritical
    from pfb.wavelets.wavelets import wavelet_setup
    from pfb.prox.prox_21m import prox_21m_numba as prox_21
    # from pfb.prox.prox_21 import prox_21
    from pfb.utils.misc import fitcleanbeam

    basename = f'{opts.output_filename}_{opts.product.upper()}'

    dds_name = f'{basename}_{opts.postfix}.dds'
    dds = xds_from_zarr(dds_name, chunks={'row':-1,
                                          'chan':-1,
                                          'x':-1,
                                          'y':-1,
                                          'x_psf':-1,
                                          'y_psf':-1,
                                          'yo2':-1})
    if opts.memory_greedy:
        dds = dask.persist(dds)[0]

    nx_psf, ny_psf = dds[0].x_psf.size, dds[0].y_psf.size
    lastsize = ny_psf

    # stitch dirty/psf in apparent scale
    print("Combining slices into cubes", file=log)
    output_type = dds[0].DIRTY.dtype
    dirty, model, residual, psf, psfhat, beam, wsums, dual = dds2cubes(
                                                               dds,
                                                               opts.nband,
                                                               apparent=False)
    wsum = np.sum(wsums)
    psf_mfs = np.sum(psf, axis=0)
    assert (psf_mfs.max() - 1.0) < 2*opts.epsilon
    dirty_mfs = np.sum(dirty, axis=0)
    if residual is None:
        residual = dirty.copy()
        residual_mfs = dirty_mfs.copy()
    else:
        residual_mfs = np.sum(residual, axis=0)

    # for intermediary results (not currently written)
    freq_out = []
    for ds in dds:
        freq_out.append(ds.freq_out)
    freq_out = np.unique(np.array(freq_out))
    nband = opts.nband
    nx = dds[0].x.size
    ny = dds[0].y.size
    ra = dds[0].ra
    dec = dds[0].dec
    radec = [ra, dec]
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)
    ref_freq = np.mean(freq_out)
    hdr_mfs = set_wcs(cell_deg, cell_deg, nx, ny, radec, ref_freq)

    # # TODO - check coordinates match
    # # Add option to interp onto coordinates?
    # if opts.mask is not None:
    #     mask = load_fits(mask, dtype=output_type).squeeze()
    #     assert mask.shape == (nx, ny)
    #     mask = mask.astype(output_type)
    #     print('Using provided fits mask', file=log)
    # else:
    #     mask = np.ones((nx, ny), dtype=output_type)

    # set up vis space Hessian
    hessopts = {}
    hessopts['cell'] = dds[0].cell_rad
    hessopts['wstack'] = opts.wstack
    hessopts['epsilon'] = opts.epsilon
    hessopts['double_accum'] = opts.double_accum
    hessopts['nthreads'] = opts.nvthreads  # nvthreads since dask parallel over band
    # always clean in apparent scale so no beam
    # mask is applied to residual after hessian application
    hess = partial(hessian_xds, xds=dds, hessopts=hessopts,
                   wsum=wsum, sigmainv=0,
                   mask=np.ones((nx, ny), dtype=output_type),
                   compute=True, use_beam=False)


    # image space hessian
    # pre-allocate arrays for doing FFT's
    xout = np.empty(dirty.shape, dtype=dirty.dtype, order='C')
    xout = make_noncritical(xout)
    xpad = np.empty(psf.shape, dtype=dirty.dtype, order='C')
    xpad = make_noncritical(xpad)
    xhat = np.empty(psfhat.shape, dtype=psfhat.dtype)
    xhat = make_noncritical(xhat)
    psf_convolve = partial(psf_convolve_cube, xpad, xhat, xout, psfhat, lastsize,
                           nthreads=opts.nvthreads*opts.nthreads_dask)  # nthreads = nvthreads*nthreads_dask because dask not involved

    if opts.hessnorm is None:
        print("Finding spectral norm of Hessian approximation", file=log)
        hessnorm, hessbeta = power_method(psf_convolve, (nband, nx, ny),
                                          tol=opts.pm_tol,
                                          maxit=opts.pm_maxit,
                                          verbosity=opts.pm_verbose,
                                          report_freq=opts.pm_report_freq)
    else:
        hessnorm = opts.hessnorm
        print(f"Using provided hessnorm of beta = {hessnorm:.3e}", file=log)

    print("Setting up dictionary", file=log)
    bases = tuple(opts.bases.split(','))
    nbasis = len(bases)
    iy, sy, ntot, nmax = wavelet_setup(
                                np.zeros((1, nx, ny), dtype=dirty.dtype),
                                bases, opts.nlevels)
    ntot = tuple(ntot)

    psiH = partial(im2coef,
                   bases=bases,
                   ntot=ntot,
                   nmax=nmax,
                   nlevels=opts.nlevels,
                   nthreads=opts.nvthreads*opts.nthreads_dask) # nthreads = nvthreads*nthreads_dask because dask not involved
    psi = partial(coef2im,
                  bases=bases,
                  ntot=ntot,
                  iy=iy,
                  sy=sy,
                  nx=nx,
                  ny=ny,
                  nthreads=opts.nvthreads*opts.nthreads_dask) # nthreads = nvthreads*nthreads_dask because dask not involved

    # get clean beam area to convert residual units during l1reweighting
    # TODO - could refine this with comparison between dirty and restored
    # if contiuing the deconvolution
    GaussPar = fitcleanbeam(psf_mfs[None], level=0.5, pixsize=1.0)[0]
    pix_per_beam = GaussPar[0]*GaussPar[1]*np.pi/4
    print(f"Number of pixels per beam estimated as {pix_per_beam}",
          file=log)

    # We do the following to set hyper-parameters in an intuitive way
    # i) convert residual units so it is comparable to model
    # ii) project residual into dual domain
    # iii) compute the rms in the space where thresholding happens
    psiHoutvar = np.zeros((nband, nbasis, nmax), dtype=dirty.dtype)
    fsel = wsums > 0
    tmp2 = residual.copy()
    tmp2[fsel] *= wsum/wsums[fsel, None, None]
    psiH(tmp2/pix_per_beam, psiHoutvar)
    rms_comps = np.std(np.sum(psiHoutvar, axis=0),
                       axis=-1)[:, None]  # preserve axes

    # TODO - load from dds if present
    if dual is None:
        dual = np.zeros((nband, nbasis, nmax), dtype=dirty.dtype)
        l1weight = np.ones((nbasis, nmax), dtype=dirty.dtype)
        reweighter = None
    else:
        if opts.l1reweight_from == 0:
            print('Initialising with L1 reweighted', file=log)
            reweighter = partial(l1reweight_func, psiH, psiHoutvar, opts.rmsfactor, rms_comps)
            l1weight = reweighter(model)
            # l1weight[l1weight < 1.0] = 0.0
        else:
            l1weight = np.ones((nbasis, nmax), dtype=dirty.dtype)
            reweighter = None


    # for generality the prox function only takes the
    # array variable and step size as inputs
    # prox21 = partial(prox_21, weight=l1weight)

    rms = np.std(residual_mfs)
    rmax = np.abs(residual_mfs).max()
    print(f"Iter 0: peak residual = {rmax:.3e}, rms = {rms:.3e}",
          file=log)
    for k in range(opts.niter):
        print('Solving for model', file=log)
        modelp = deepcopy(model)
        data = residual + psf_convolve(model)
        grad21 = lambda x: psf_convolve(x) - data
        model, dual = primal_dual(model,
                                  dual,
                                  opts.rmsfactor*rms,
                                  psi,
                                  psiH,
                                  hessnorm,
                                  prox_21,
                                  l1weight,
                                  reweighter,
                                  grad21,
                                  nu=nbasis,
                                  positivity=opts.positivity,
                                  tol=opts.pd_tol,
                                  maxit=opts.pd_maxit,
                                  verbosity=opts.pd_verbose,
                                  report_freq=opts.pd_report_freq,
                                  gamma=opts.gamma)

        save_fits(np.mean(model, axis=0),
                  basename + f'_{opts.postfix}_model_{k+1}.fits',
                  hdr_mfs)

        print("Getting residual", file=log)
        convimage = hess(model)
        ne.evaluate('dirty - convimage', out=residual,
                    casting='same_kind')
        ne.evaluate('sum(residual, axis=0)', out=residual_mfs,
                    casting='same_kind')

        save_fits(residual_mfs,
                  basename + f'_{opts.postfix}_residual_{k+1}.fits',
                  hdr_mfs)

        rms = np.std(residual_mfs)
        rmax = np.abs(residual_mfs).max()
        eps = np.linalg.norm(model - modelp)/np.linalg.norm(model)

        print(f"Iter {k+1}: peak residual = {rmax:.3e}, "
              f"rms = {rms:.3e}, eps = {eps:.3e}",
              file=log)

        if k+1 >= opts.l1reweight_from:
            print('Computing L1 weights', file=log)
            # convert residual units so it is comparable to model
            tmp2[fsel] = residual[fsel] * wsum/wsums[fsel, None, None]
            psiH(tmp2/pix_per_beam, psiHoutvar)
            rms_comps = np.std(np.sum(psiHoutvar, axis=0),
                               axis=-1)[:, None]  # preserve axes
            # we redefine the reweighter here since the rms has changed
            reweighter = partial(l1reweight_func, psiH, psiHoutvar, opts.rmsfactor, rms_comps)
            l1weight = reweighter(model)
            # l1weight[l1weight < 1.0] = 0.0
            # prox21 = partial(prox_21, weight=l1weight, axis=0)

        print("Updating results", file=log)
        dds_out = []
        for ds in dds:
            b = ds.bandid
            r = da.from_array(residual[b]*wsum)
            m = da.from_array(model[b])
            d = da.from_array(dual[b])
            ds_out = ds.assign(**{'RESIDUAL': (('x', 'y'), r),
                                  'MODEL': (('x', 'y'), m),
                                  'DUAL': (('c', 'n'), d)})
            dds_out.append(ds_out)
        writes = xds_to_zarr(dds_out, dds_name,
                             columns=('RESIDUAL', 'MODEL', 'DUAL'),
                             rechunk=True)
        dask.compute(writes)

        if eps < opts.tol:
            print(f"Converged after {k+1} iterations.", file=log)
            break
        # if rmax <= threshold:
        #     print("Terminating because final threshold has been reached",
        #           file=log)
        #     break

    dds = xds_from_zarr(dds_name, chunks={'x': -1, 'y': -1})

    # convert to fits files
    fitsout = []
    if opts.fits_mfs:
        fitsout.append(dds2fits_mfs(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
        fitsout.append(dds2fits_mfs(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if opts.fits_cubes:
        fitsout.append(dds2fits(dds, 'RESIDUAL', f'{basename}_{opts.postfix}', norm_wsum=True))
        fitsout.append(dds2fits(dds, 'MODEL', f'{basename}_{opts.postfix}', norm_wsum=False))

    if len(fitsout):
        print("Writing fits", file=log)
        dask.compute(fitsout)

    print("All done here.", file=log)


# def _spotless_dist(**kw):
#     opts = OmegaConf.create(kw)
#     OmegaConf.set_struct(opts, True)

#     import numpy as np
#     import xarray as xr
#     import dask
#     import dask.array as da
#     from distributed import Client, wait, get_client, as_completed
#     from pfb.opt.power_method import power_method_dist as power_method
#     from pfb.opt.pcg import pcg_dist as pcg
#     from pfb.opt.primal_dual import primal_dual_dist as primal_dual
#     from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
#     from pfb.utils.fits import load_fits, dds2fits, dds2fits_mfs
#     from pfb.utils.dist import (get_resid_and_stats, accum_wsums,
#                                 compute_residual, update_results,
#                                 get_epsb, l1reweight, get_cbeam_area,
#                                 set_wsum, get_eps, set_l1weight)
#     from pfb.operators.hessian import hessian_psf_slice
#     from pfb.operators.psi import im2coef_dist as im2coef
#     from pfb.operators.psi import coef2im_dist as coef2im
#     from pfb.prox.prox_21m import prox_21m
#     from pfb.prox.prox_21 import prox_21
#     from pfb.wavelets.wavelets import wavelet_setup
#     import pywt
#     from copy import deepcopy
#     from operator import getitem
#     from itertools import cycle

#     basename = f'{opts.output_filename}_{opts.product.upper()}'
#     dds_name = f'{basename}_{opts.postfix}.dds'

#     client = get_client()
#     names = [w['name'] for w in client.scheduler_info()['workers'].values()]

#     dds = xds_from_zarr(dds_name, chunks={'row':-1,
#                                           'chan':-1,
#                                           'x':-1,
#                                           'y':-1,
#                                           'x_psf':-1,
#                                           'y_psf':-1,
#                                           'yo2':-1,
#                                           'b':-1,
#                                           'c':-1})

#     real_type = dds[0].DIRTY.dtype
#     cell_rad = dds[0].cell_rad
#     complex_type = np.result_type(dirty.dtype, np.complex64)

#     nband = len(dds)
#     nx, ny = dds[0].nx, dds[0].ny
#     nx_psf, ny_psf = dds[0].nx_psf, dds[0].ny_psf
#     nx_psf, nyo2_psf = dds[0].PSFHAT.shape
#     npad_xl = (nx_psf - nx)//2
#     npad_xr = nx_psf - nx - npad_xl
#     npad_yl = (ny_psf - ny)//2
#     npad_yr = ny_psf - ny - npad_yl
#     psf_padding = ((npad_xl, npad_xr), (npad_yl, npad_yr))
#     unpad_x = slice(npad_xl, -npad_xr)
#     unpad_y = slice(npad_yl, -npad_yr)
#     lastsize = ny + np.sum(psf_padding[-1])

#     # dictionary setup
#     print("Setting up dictionary", file=log)
#     bases = tuple(opts.bases.split(','))
#     nbasis = len(bases)
#     iy, sy, ntot, nmax = wavelet_setup(
#                                 np.zeros((1, nx, ny), dtype=dirty.dtype),
#                                 bases, opts.nlevels)
#     ntot = tuple(ntot)

#     psiH = partial(im2coef,
#                     bases=bases,
#                     ntot=ntot,
#                     nmax=nmax,
#                     nlevels=opts.nlevels)
#     # avoid pickling dumba Dict
#     psi = partial(coef2im,
#                     bases=bases,
#                     ntot=ntot,
#                     iy=None,  #dict(iy),
#                     sy=None,  #dict(sy),
#                     nx=nx,
#                     ny=ny)

#     print('Scattering data', file=log)
#     Afs = {}
#     for ds, i in zip(dds, cycle(range(opts.nworkers))):
#         Af = client.submit(hessian_psf_slice, ds, nbasis, nmax,
#                            opts.nvthreads,
#                            opts.sigmainv,
#                            workers={names[i]})
#         Afs[names[i]] = Af

#     # wait expects a list
#     wait(list(Afs.values()))

#     # assumed constant
#     wsum = client.submit(accum_wsums, Afs, workers={names[0]}).result()
#     pix_per_beam = client.submit(get_cbeam_area, Afs, wsum, workers={names[0]}).result()

#     for wid, A in Afs.items():
#         client.submit(set_wsum, A, wsum,
#                       pure=False,
#                       workers={wid})

#     try:
#         l1ds = xds_from_zarr(f'{dds_name}::L1WEIGHT', chunks={'b':-1,'c':-1})
#         if 'L1WEIGHT' in l1ds:
#             l1weightfs = client.submit(set_l1weight, l1ds,
#                                       workers={names[0]})
#         else:
#             raise
#     except Exception as e:
#         print(f'Did not find l1weights at {dds_name}/L1WEIGHT. '
#               'Initialising to unity', file=log)
#         l1weightfs = client.submit(np.ones, (nbasis, nmax), dtype=dirty.dtype,
#                                  workers={names[0]})

#     if opts.hessnorm is None:
#         print('Getting spectral norm of Hessian approximation', file=log)
#         hessnorm = power_method(Afs, nx, ny, nband)
#     else:
#         hessnorm = opts.hessnorm
#     print(f'hessnorm = {hessnorm:.3e}', file=log)

#     # future contains mfs residual and stats
#     residf = client.submit(get_resid_and_stats, Afs, wsum,
#                            workers={names[0]})
#     residual_mfs = client.submit(getitem, residf, 0, workers={names[0]})
#     rms = client.submit(getitem, residf, 1, workers={names[0]}).result()
#     rmax = client.submit(getitem, residf, 2, workers={names[0]}).result()
#     print(f"It {0}: max resid = {rmax:.3e}, rms = {rms:.3e}", file=log)
#     for k in range(opts.niter):
#         print('Solving for update', file=log)
#         xfs = {}
#         for wid, A in Afs.items():
#             xf = client.submit(pcg, A,
#                                opts.cg_maxit,
#                                opts.cg_minit,
#                                opts.cg_tol,
#                                opts.sigmainv,
#                                workers={wid})
#             xfs[wid] = xf

#         # wait expects a list
#         wait(list(xfs.values()))

#         print('Solving for model', file=log)
#         modelfs, dualfs = primal_dual(
#                             Afs, xfs,
#                             psi, psiH,
#                             opts.rmsfactor*rms,
#                             hessnorm,
#                             l1weightfs,
#                             nu=len(bases),
#                             tol=opts.pd_tol,
#                             maxit=opts.pd_maxit,
#                             positivity=opts.positivity,
#                             gamma=opts.gamma,
#                             verbosity=opts.pd_verbose)

#         print('Computing residual', file=log)
#         residfs = {}
#         for wid, A in Afs.items():
#             residf = client.submit(compute_residual, A, modelfs[wid],
#                                    workers={wid})
#             residfs[wid] = residf

#         wait(list(residfs.values()))

#         # l1reweighting
#         if k+1 >= opts.l1reweight_from:
#             print('L1 reweighting', file=log)
#             l1weightfs = client.submit(l1reweight, modelfs, residfs, l1weightfs,
#                                        psiH, wsum, pix_per_beam, workers=names[0])

#         # dump results so we can continue from if needs be
#         print('Updating results', file=log)
#         ddsfs = {}
#         modelpfs = {}
#         for i, (wid, A) in enumerate(Afs.items()):
#             modelpfs[wid] = client.submit(getattr, A, 'model', workers={wid})
#             dsf = client.submit(update_results, A, dds[i], modelfs[i], dualfs[i], residfs[i],
#                                 pure=False,
#                                 workers={wid})
#             ddsfs[wid] = dsf

#         dds = []
#         for f in as_completed(list(ddsfs.values())):
#             dds.append(f.result())
#         writes = xds_to_zarr(dds, dds_name,
#                              columns=('MODEL','DUAL','RESIDUAL'),
#                              rechunk=True)
#         l1weight = da.from_array(l1weightfs.result(), chunks=(1, 4096**2))
#         dvars = {}
#         dvars['L1WEIGHT'] = (('b','c'), l1weight)
#         l1ds = xr.Dataset(dvars)
#         l1writes = xds_to_zarr(l1ds, f'{dds_name}::L1WEIGHT')
#         client.compute(writes, l1writes)

#         residf = client.submit(get_resid_and_stats, Afs, wsum,
#                                workers={names[0]})
#         residual_mfs = client.submit(getitem, residf, 0, workers={names[0]})
#         rms = client.submit(getitem, residf, 1, workers={names[0]}).result()
#         rmax = client.submit(getitem, residf, 2, workers={names[0]}).result()
#         eps_num = []
#         eps_den = []
#         for wid in Afs.keys():
#             fut = client.submit(get_epsb, modelpfs[wid], modelfs[wid], workers={wid})
#             eps_num.append(client.submit(getitem, fut, 0, workers={wid}))
#             eps_den.append(client.submit(getitem, fut, 1, workers={wid}))

#         eps = client.submit(get_eps, eps_num, eps_den, workers={names[0]}).result()
#         print(f"It {k+1}: max resid = {rmax:.3e}, rms = {rms:.3e}, eps = {eps:.3e}",
#               file=log)


#         if eps < opts.tol:
#             break

#     # convert to fits files
#     fitsout = []
#     if opts.fits_mfs:
#         fitsout.append(dds2fits_mfs(dds, 'RESIDUAL', basename, norm_wsum=True))
#         fitsout.append(dds2fits_mfs(dds, 'MODEL', basename, norm_wsum=False))

#     if opts.fits_cubes:
#         fitsout.append(dds2fits(dds, 'RESIDUAL', basename, norm_wsum=True))
#         fitsout.append(dds2fits(dds, 'MODEL', basename, norm_wsum=False))

#     if len(fitsout):
#         print("Writing fits", file=log)
#         dask.compute(fitsout)

#     client.close()

#     print("All done here.", file=log)
#     return


# def Idty(x):
#     return x



"""
Dask wrappers around the wgridder. These operators are per band
because we are not guaranteed that each imaging band has the same
number of rows after BDA.
"""

import numpy as np
import dask
import dask.array as da
from ducc0.wgridder.experimental import vis2dirty, dirty2vis
from ducc0.fft import c2r, r2c
from africanus.constants import c as lightspeed
from quartical.utils.dask import Blocker
from pfb.utils.weighting import _counts_to_weights
iFs = np.fft.ifftshift
Fs = np.fft.fftshift

def vis2im(uvw=None,
           freq=None,
           vis=None,
           wgt=None,
           nx=None, ny=None,
           cellx=None, celly=None,
           nthreads=None,
           epsilon=None,
           mask=None,
           do_wgridding=True,
           flip_v=False,
           x0=0, y0=0,
           precision='single',
           divide_by_n=False,
           sigma_min=1.1, sigma_max=2.6,
           double_precision_accumulation=True):

    if precision.lower() == 'single':
        real_type = np.float32
    elif precision.lower() == 'double':
        real_type = np.float64

    if wgt is not None:
        wgt_out = 'rf'
    else:
        wgt_out = None

    if mask is not None:
        mask_out = 'rf'
    else:
        mask_out = None

    return da.blockwise(_vis2im, 'xy',
                        uvw, 'r3',
                        freq, 'f',
                        vis, 'rf',
                        wgt, wgt_out,
                        mask, mask_out,
                        nx, None,
                        ny, None,
                        cellx, None,
                        celly, None,
                        x0, None,
                        y0, None,
                        epsilon, None,
                        precision, None,
                        flip_v, None,
                        do_wgridding, None,
                        divide_by_n, None,
                        nthreads, None,
                        sigma_min, None,
                        sigma_max, None,
                        double_precision_accumulation, None,
                        new_axes={'x':nx, 'y': ny},
                        dtype=real_type)

def _vis2im(uvw,
            freq,
            vis,
            wgt,
            mask,
            nx, ny,
            cellx, celly,
            x0, y0,
            epsilon,
            precision,
            flip_v,
            do_wgridding,
            divide_by_n,
            nthreads,
            sigma_min,
            sigma_max,
            double_precision_accumulation):

    if wgt is not None:
        wgt_out = wgt[0][0]
    else:
        wgt_out = None

    if mask is not None:
        mask_out = mask[0][0]
    else:
        mask_out = None

    return _vis2im_impl(uvw[0][0],
                        freq[0],
                        vis[0][0],
                        wgt_out,
                        mask_out,
                        nx, ny,
                        cellx, celly,
                        x0, y0,
                        epsilon,
                        precision,
                        flip_v,
                        do_wgridding,
                        divide_by_n,
                        nthreads,
                        sigma_min, sigma_max,
                        double_precision_accumulation)

def _vis2im_impl(uvw,
                 freq,
                 vis,
                 wgt,
                 mask,
                 nx, ny,
                 cellx, celly,
                 x0, y0,
                 epsilon,
                 precision,
                 flip_v,
                 do_wgridding,
                 divide_by_n,
                 nthreads,
                 sigma_min, sigma_max,
                 double_precision_accumulation):
    uvw = np.require(uvw, dtype=np.float64)
    freq = np.require(freq, np.float64)

    if precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128

    vis = np.require(vis, dtype=complex_type)

    if wgt is not None:
        wgt = np.require(wgt, dtype=real_type)

    if mask is not None:
        mask = np.require(mask, dtype=np.uint8)

    return vis2dirty(uvw=uvw,
                     freq=freq,
                     vis=vis,
                     wgt=wgt,
                     mask=mask,
                     npix_x=nx, npix_y=ny,
                     pixsize_x=cellx, pixsize_y=celly,
                     center_x=x0, center_y=y0,
                     epsilon=epsilon,
                     flip_v=flip_v,
                     do_wgridding=do_wgridding,
                     divide_by_n=divide_by_n,
                     nthreads=nthreads,
                     sigma_min=sigma_min, sigma_max=sigma_max,
                     double_precision_accumulation=double_precision_accumulation)



def im2vis(uvw=None,
           freq=None,
           image=None,
           wgt=None,
           mask=None,
           cellx=None,
           celly=None,
           nthreads=1,
           epsilon=1e-7,
           do_wgridding=True,
           flip_v=False,
           x0=0, y0=0,
           precision='single',
           divide_by_n=False,
           sigma_min=1.1,
           sigma_max=2.6):

    if precision.lower() == 'single':
        complex_type = np.float32
    elif precision.lower() == 'double':
        complex_type = np.float64

    if wgt is not None:
        wgt_out = 'rf'
    else:
        wgt_out = None

    if mask is not None:
        mask_out = 'rf'
    else:
        mask_out = None

    return da.blockwise(_im2vis, 'rf',
                        uvw, 'r3',
                        freq, 'f',
                        image, 'xy',
                        wgt, wgt_out,
                        mask, mask_out,
                        cellx, None,
                        celly, None,
                        x0, None,
                        y0, None,
                        epsilon, None,
                        flip_v, None,
                        do_wgridding, None,
                        precision, None,
                        divide_by_n, None,
                        nthreads, None,
                        sigma_min, None,
                        sigma_max, None,
                        dtype=complex_type)

def _im2vis(uvw,
           freq,
           image,
           wgt,
           mask,
           cellx,
           celly,
           x0=0, y0=0,
           epsilon=1e-7,
           flip_v=False,
           do_wgridding=True,
           precision='single',
           divide_by_n=False,
           nthreads=1,
           sigma_min=1.1,
           sigma_max=2.6):

    return _im2vis_impl(uvw[0],
                        freq,
                        image[0][0],
                        wgt,
                        mask,
                        cellx, celly,
                        x0, y0,
                        epsilon,
                        flip_v,
                        do_wgridding,
                        precision,
                        divide_by_n,
                        nthreads,
                        sigma_min,
                        sigma_max)



def _im2vis_impl(uvw,
                 freq,
                 image,
                 wgt,
                 mask,
                 cellx,
                 celly,
                 x0=0, y0=0,
                 epsilon=1e-7,
                 flip_v=False,
                 do_wgridding=True,
                 precision='single',
                 divide_by_n=False,
                 nthreads=1,
                 sigma_min=1.1,
                 sigma_max=2.6):
    uvw = np.require(uvw, dtype=np.float64)
    freq = np.require(freq, np.float64)
    if precision.lower() == 'single':
        real_type = np.float32
        complex_type = np.complex64
    elif precision.lower() == 'double':
        real_type = np.float64
        complex_type = np.complex128
    image = np.require(image, dtype=real_type)
    if wgt is not None:
        wgt = np.require(wgt, dtype=real_type)
    if mask is not None:
        mask = np.require(mask, dtype=np.uint8)
    return dirty2vis(uvw=uvw, freq=freq, dirty=image, wgt=wgt, mask=mask,
                     pixsize_x=cellx, pixsize_y=celly,
                     center_x=x0, center_y=y0, epsilon=epsilon, flip_v=flip_v,
                     do_wgridding=do_wgridding, divide_by_n=divide_by_n,
                     nthreads=nthreads, sigma_min=sigma_min,
                     sigma_max=sigma_max)



def _loc2psf_vis_impl(uvw,
                      freq,
                      cell,
                      x0=0,
                      y0=0,
                      wstack=True,
                      epsilon=1e-7,
                      nthreads=1,
                      precision='single',
                      divide_by_n=False,
                      convention='CASA'):
    if x0 or y0:
        # LB - what is wrong with this?
        # n = np.sqrt(1 - x0**2 - y0**2)
        # if convention.upper() == 'CASA':
        #     freqfactor = -2j*np.pi*freq[None, :]/lightspeed
        # else:
        #     freqfactor = 2j*np.pi*freq[None, :]/lightspeed
        # psf_vis = np.exp(freqfactor*(uvw[:, 0:1]*x0 +
        #                              uvw[:, 1:2]*y0 +
        #                              uvw[:, 2:]*(n-1)))
        # if divide_by_n:
        #     psf_vis /= n
        real_type = np.float32 if precision == 'single' else np.float64
        x = np.zeros((128,128), dtype=real_type)
        x[64,64] = 1.0
        psf_vis = dirty2vis(uvw=uvw,
                            freq=freq,
                            dirty=x,
                            pixsize_x=cell,
                            pixsize_y=cell,
                            center_x=x0,
                            center_y=y0,
                            epsilon=1e-7,
                            do_wgridding=wstack,
                            nthreads=nthreads,
                            divide_by_n=divide_by_n)

    else:
        nrow, _ = uvw.shape
        nchan = freq.size
        dtype = np.complex64 if precision == 'single' else np.complex128
        psf_vis = np.ones((nrow, nchan), dtype=dtype)
    return psf_vis


def _loc2psf_vis(uvw,
                 freq,
                 cell,
                 x0=0,
                 y0=0,
                 wstack=True,
                 epsilon=1e-7,
                 nthreads=1,
                 precision='single',
                 divide_by_n=False,
                 convention='CASA'):
    return _loc2psf_vis_impl(uvw[0],
                             freq,
                             cell,
                             x0,
                             y0,
                             wstack,
                             epsilon,
                             nthreads,
                             precision,
                             divide_by_n,
                             convention)


def loc2psf_vis(uvw,
                freq,
                cell,
                x0=0,
                y0=0,
                wstack=True,
                epsilon=1e-7,
                nthreads=1,
                precision='single',
                divide_by_n=False,
                convention='CASA'):
    return da.blockwise(_loc2psf_vis, 'rf',
                        uvw, 'r3',
                        freq, 'f',
                        cell, None,
                        x0, None,
                        y0, None,
                        wstack, None,
                        epsilon, None,
                        nthreads, None,
                        precision, None,
                        divide_by_n, None,
                        convention, None,
                        dtype=np.complex64 if precision == 'single' else np.complex128)


def comps2vis(uvw,
              utime,
              freq,
              rbin_idx, rbin_cnts,
              tbin_idx, tbin_cnts,
              fbin_idx, fbin_cnts,
              comps,
              Ix, Iy,
              modelf,
              tfunc,
              ffunc,
              nx, ny,
              cellx, celly,
              x0=0, y0=0,
              epsilon=1e-7,
              nthreads=1,
              wstack=True,
              divide_by_n=False,
              ncorr_out=4):

    # determine output type
    complex_type = da.result_type(comps, np.complex64)

    return da.blockwise(_comps2vis, 'rfc',
                        uvw, 'r3',
                        utime, 'r',
                        freq, 'f',
                        rbin_idx, 'r',
                        rbin_cnts, 'r',
                        tbin_idx, 'r',
                        tbin_cnts, 'r',
                        fbin_idx, 'f',
                        fbin_cnts, 'f',
                        comps, None,
                        Ix, None,
                        Iy, None,
                        modelf, None,
                        tfunc, None,
                        ffunc, None,
                        nx, None,
                        ny, None,
                        cellx, None,
                        celly, None,
                        x0, None,
                        y0, None,
                        epsilon, None,
                        nthreads, None,
                        wstack, None,
                        divide_by_n, None,
                        ncorr_out, None,
                        new_axes={'c': ncorr_out},
                        # it should be pulling these from uvw and freq so shouldn't need this?
                        adjust_chunks={'r': uvw.chunks[0]},
                        dtype=complex_type,
                        align_arrays=False)


def _comps2vis(uvw,
                utime,
                freq,
                rbin_idx, rbin_cnts,
                tbin_idx, tbin_cnts,
                fbin_idx, fbin_cnts,
                comps,
                Ix, Iy,
                modelf,
                tfunc,
                ffunc,
                nx, ny,
                cellx, celly,
                x0=0, y0=0,
                epsilon=1e-7,
                nthreads=1,
                wstack=True,
                divide_by_n=False,
                ncorr_out=4):
    return _comps2vis_impl(uvw[0],
                           utime,
                           freq,
                           rbin_idx, rbin_cnts,
                           tbin_idx, tbin_cnts,
                           fbin_idx, fbin_cnts,
                           comps,
                           Ix, Iy,
                           modelf,
                           tfunc,
                           ffunc,
                           nx, ny,
                           cellx, celly,
                           x0=x0, y0=y0,
                           epsilon=epsilon,
                           nthreads=nthreads,
                           wstack=wstack,
                           divide_by_n=divide_by_n,
                           ncorr_out=ncorr_out)



def _comps2vis_impl(uvw,
                    utime,
                    freq,
                    rbin_idx, rbin_cnts,
                    tbin_idx, tbin_cnts,
                    fbin_idx, fbin_cnts,
                    comps,
                    Ix, Iy,
                    modelf,
                    tfunc,
                    ffunc,
                    nx, ny,
                    cellx, celly,
                    x0=0, y0=0,
                    epsilon=1e-7,
                    nthreads=1,
                    wstack=True,
                    divide_by_n=False,
                    ncorr_out=4):
    # adjust for chunking
    # need a copy here if using multiple row chunks
    rbin_idx2 = rbin_idx - rbin_idx.min()
    tbin_idx2 = tbin_idx - tbin_idx.min()
    fbin_idx2 = fbin_idx - fbin_idx.min()

    # currently not interpolating in time
    ntime = tbin_idx.size
    nband = fbin_idx.size

    nrow = uvw.shape[0]
    nchan = freq.size
    vis = np.zeros((nrow, nchan, ncorr_out), dtype=np.result_type(comps, np.complex64))
    for t in range(ntime):
        indt = slice(tbin_idx2[t], tbin_idx2[t] + tbin_cnts[t])
        # TODO - clean up this logic. row_mapping holds the number of rows per
        # utime and there are multiple utimes per row chunk
        indr = slice(rbin_idx2[indt][0], rbin_idx2[indt][-1] + rbin_cnts[indt][-1])
        for b in range(nband):
            indf = slice(fbin_idx2[b], fbin_idx2[b] + fbin_cnts[b])
            # render components to image
            # we want to do this on each worker
            tout = tfunc(np.mean(utime[indt]))
            fout = ffunc(np.mean(freq[indf]))
            image = np.zeros((nx, ny), dtype=comps.dtype)
            image[Ix, Iy] = modelf(tout, fout, *comps[:, :])  # too magical?
            vis[indr, indf, 0] = dirty2vis(uvw=uvw,
                                           freq=freq[indf],
                                           dirty=image,
                                           pixsize_x=cellx, pixsize_y=celly,
                                           center_x=x0, center_y=y0,
                                           epsilon=epsilon,
                                           do_wgridding=wstack,
                                           divide_by_n=divide_by_n,
                                           nthreads=nthreads)
            if ncorr_out > 1:
                vis[indr, indf, -1] = vis[indr, indf, 0]
    return vis


def image_data_products(uvw,
                        freq,
                        vis,
                        wgt,
                        mask,
                        counts,
                        nx, ny,
                        nx_psf, ny_psf,
                        cellx, celly,
                        model=None,
                        robustness=None,
                        x0=0.0, y0=0.0,
                        nthreads=1,
                        epsilon=1e-7,
                        do_wgridding=True,
                        double_accum=True,
                        # divide_by_n=False,
                        l2reweight_dof=None,
                        do_dirty=True,
                        do_psf=True,
                        do_weight=True,
                        do_residual=False):
    '''
    Function to compute image space data products in one go
        dirty
        psf
        psfhat
        imweight
        residual
        wsum
    '''
    out_dict = {}

    if model is None:
        if l2reweight_dof:
            raise ValueError('Requested l2 reweight but no model passed in. '
                             'Perhaps transfer model from somewhere?')
    else:
        # do not apply weights in this direction
        model_vis = dirty2vis(
            uvw=uvw,
            freq=freq,
            dirty=model,
            pixsize_x=cellx,
            pixsize_y=celly,
            center_x=x0,
            center_y=y0,
            epsilon=epsilon,
            do_wgridding=do_wgridding,
            flip_v=False,
            nthreads=nthreads,
            divide_by_n=False,
            sigma_min=1.1, sigma_max=3.0)

        residual_vis = vis - model_vis
        # apply mask to both
        residual_vis *= mask

    if l2reweight_dof:
        ressq = (residual_vis*residual_vis.conj()).real
        wcount = mask.sum()
        if wcount:
            ovar = ressq.sum()/wcount
            wgt = (dof + 1)/(dof + ressq/ovar)/ovar

    # we usually want to re-evaluate this since the robustness may change
    if robustness is not None:
        if counts is None:
            raise ValueError('counts are None but robustness specified. '
                             'This is probably a bug!')
        imwgt = _counts_to_weights(
            counts,
            uvw,
            freq,
            nx, ny,
            cellx, celly,
            robustness)

        wgt *= imwgt

    if do_weight:
        out_dict['WEIGHT'] = wgt

    wsum = wgt[mask.astype(bool)].sum()
    out_dict['WSUM'] = np.atleast_1d(wsum)

    dirty = vis2dirty(
        uvw=uvw,
        freq=freq,
        vis=vis,
        wgt=wgt,
        mask=mask,
        npix_x=nx, npix_y=ny,
        pixsize_x=cellx, pixsize_y=celly,
        center_x=x0, center_y=y0,
        epsilon=epsilon,
        flip_v=False,  # hardcoded for now
        do_wgridding=do_wgridding,
        divide_by_n=False,  # hardcoded for now
        nthreads=nthreads,
        sigma_min=1.1, sigma_max=3.0,
        double_precision_accumulation=double_accum)
    out_dict['DIRTY'] = dirty

    if do_psf:
        if x0 or y0:
        # LB - what is wrong with this?
        # n = np.sqrt(1 - x0**2 - y0**2)
        # if convention.upper() == 'CASA':
        #     freqfactor = -2j*np.pi*freq[None, :]/lightspeed
        # else:
        #     freqfactor = 2j*np.pi*freq[None, :]/lightspeed
        # psf_vis = np.exp(freqfactor*(uvw[:, 0:1]*x0 +
        #                              uvw[:, 1:2]*y0 +
        #                              uvw[:, 2:]*(n-1)))
        # if divide_by_n:
        #     psf_vis /= n
            x = np.zeros((128,128), dtype=wgt.dtype)
            x[64,64] = 1.0
            psf_vis = dirty2vis(
                uvw=uvw,
                freq=freq,
                dirty=x,
                pixsize_x=cellx,
                pixsize_y=celly,
                center_x=x0,
                center_y=y0,
                epsilon=1e-7,
                do_wgridding=wstack,
                nthreads=nthreads,
                divide_by_n=False,
                flip_v=False,  # hardcoded for now
                sigma_min=1.1, sigma_max=3.0)

        else:
            nrow, _ = uvw.shape
            nchan = freq.size
            psf_vis = np.ones((nrow, nchan), dtype=vis.dtype)

        psf = vis2dirty(
            uvw=uvw,
            freq=freq,
            vis=psf_vis,
            wgt=wgt,
            mask=mask,
            npix_x=nx_psf, npix_y=ny_psf,
            pixsize_x=cellx, pixsize_y=celly,
            center_x=x0, center_y=y0,
            epsilon=epsilon,
            flip_v=False,  # hardcoded for now
            do_wgridding=do_wgridding,
            divide_by_n=False,  # hardcoded for now
            nthreads=nthreads,
            sigma_min=1.1, sigma_max=3.0,
            double_precision_accumulation=double_accum)

        # get FT of psf
        psfhat = r2c(iFs(psf, axes=(0, 1)), axes=(0, 1),
                     nthreads=nthreads,
                     forward=True, inorm=0)

        out_dict["PSF"] = psf
        out_dict["PSFHAT"] = psfhat


    if do_residual and model is not None:
        residual = vis2dirty(
            uvw=uvw,
            freq=freq,
            vis=residual_vis,
            wgt=wgt,
            mask=mask,
            npix_x=nx, npix_y=ny,
            pixsize_x=cellx, pixsize_y=celly,
            center_x=x0, center_y=y0,
            epsilon=epsilon,
            flip_v=False,  # hardcoded for now
            do_wgridding=do_wgridding,
            divide_by_n=False,  # hardcoded for now
            nthreads=nthreads,
            sigma_min=1.1, sigma_max=3.0,
            double_precision_accumulation=double_accum)

        out_dict['RESIDUAL'] = residual

    return out_dict



# flake8: noqa
from contextlib import ExitStack
from pfb.workers.main import cli
import click
from omegaconf import OmegaConf
import pyscilog
pyscilog.init('pfb')
log = pyscilog.get_logger('MODEL2COMPS')


from scabha.schema_utils import clickify_parameters
from pfb.parser.schemas import schema

# create default parameters from schema
defaults = {}
for key in schema.model2comps["inputs"].keys():
    defaults[key] = schema.model2comps["inputs"][key]["default"]

@cli.command(context_settings={'show_default': True})
@clickify_parameters(schema.model2comps)
def model2comps(**kw):
    '''
    Convert model to parameters to make it possible to evaluate it at
    an arbitrary (t, nu)
    '''
    defaults.update(kw)
    opts = OmegaConf.create(defaults)
    import time
    timestamp = time.strftime("%Y%m%d-%H%M%S")
    pyscilog.log_to_file(f'model2comps_{timestamp}.log')

    OmegaConf.set_struct(opts, True)

    if opts.product.upper() not in ["I"]:
                                    # , "Q", "U", "V", "XX", "YX", "XY",
                                    # "YY", "RR", "RL", "LR", "LL"]:
        raise NotImplementedError(f"Product {opts.product} not yet supported")

    with ExitStack() as stack:
        from pfb import set_client
        opts = set_client(opts, stack, log, scheduler=opts.scheduler)

        # TODO - prettier config printing
        print('Input Options:', file=log)
        for key in opts.keys():
            print('     %25s = %s' % (key, opts[key]), file=log)

        return _model2comps(**opts)

def _model2comps(**kw):
    opts = OmegaConf.create(kw)
    OmegaConf.set_struct(opts, True)

    import os
    import numpy as np
    from pfb.utils.misc import construct_mappings
    import dask
    from dask.distributed import performance_report
    from daskms.experimental.zarr import xds_from_zarr, xds_to_zarr
    import dask.array as da
    from africanus.constants import c as lightspeed
    from pfb.utils.fits import load_fits, data_from_header
    from astropy.io import fits
    from pfb.utils.misc import compute_context, fit_image_cube
    import xarray as xr

    basename = f'{opts.output_filename}_{opts.product.upper()}'
    dds_name = f'{basename}_{opts.postfix}.dds'
    coeff_name = f'{basename}_{opts.postfix}_{opts.model_name.lower()}.mds'

    if os.path.isdir(coeff_name):
        if opts.overwrite:
            print(f'Removing {coeff_name}', file=log)
            import shutil
            shutil.rmtree(coeff_name)

    dds = xds_from_zarr(dds_name,
                        chunks={'x':-1,
                                'y':-1})
    cell_rad = dds[0].cell_rad
    cell_deg = np.rad2deg(cell_rad)

    # get cube info
    nband = 0
    ntime = 0
    mfreqs = []
    mtimes = []
    for ds in dds:
        nband = np.maximum(ds.bandid+1, nband)
        ntime = np.maximum(ds.timeid+1, ntime)
        mfreqs.append(ds.freq_out)
        mtimes.append(ds.time_out)
    mfreqs = np.unique(np.array(mfreqs))
    mtimes = np.unique(np.array(mtimes))
    assert mfreqs.size == nband
    assert mtimes.size == ntime
    assert len(dds) == nband*ntime

    # stack cube
    nx = dds[0].x.size
    ny = dds[0].y.size
    x0 = dds[0].x0
    y0 = dds[0].y0

    # TODO - do this as a dask graph
    # model = [da.zeros((nx, ny)) for _ in range(opts.nband)]
    # wsums = [da.zeros(1) for _ in range(opts.nband)]
    model = np.zeros((ntime, nband, nx, ny), dtype=np.float64)
    wsums = np.zeros((ntime, nband), dtype=np.float64)
    for ds in dds:
        b = ds.bandid
        t = ds.timeid
        model[t, b] = getattr(ds, opts.model_name).values
        wsums[t, b] += ds.WSUM.values[0]

    if not opts.use_wsum:
        wsums[...] = 1.0

    if opts.min_val is not None:
        model = np.where(model >= opts.min_val, model, 0.0)

    if not np.any(model):
        raise ValueError(f'Model is empty or has no components above {opts.min_val}')
    radec = (dds[0].ra, dds[0].dec)

    coeffs, Ix, Iy, expr, params, texpr, fexpr = \
        fit_image_cube(mtimes, mfreqs, model, wgt=wsums,
                       nbasisf=opts.nbasisf, method=opts.fit_mode)

    # save interpolated dataset
    data_vars = {
        'coefficients': (('params', 'comps'), coeffs),
    }
    coords = {
        'location_x': (('location_x',), Ix),
        'location_y': (('location_y',), Iy),
        # 'shape_x':,
        'params': (('params',), params),  # already converted to list
        'times': (('times',), mtimes),  # to allow rendering to original grid
        'freqs': (('freqs',), mfreqs)
    }
    attrs = {
        'spec': 'genesis',
        'cell_rad_x': cell_rad,
        'cell_rad_y': cell_rad,
        'npix_x': nx,
        'npix_y': ny,
        'texpr': texpr,
        'fexpr': fexpr,
        'center_x': x0,
        'center_y': y0,
        'ra': dds[0].ra,
        'dec': dds[0].dec,
        'stokes': opts.product,  # I,Q,U,V, IQ/IV, IQUV
        'parametrisation': expr  # already converted to str
    }


    coeff_dataset = xr.Dataset(data_vars=data_vars,
                               coords=coords,
                               attrs=attrs)
    writes = xds_to_zarr(coeff_dataset,
                         coeff_name,
                         columns='ALL')
    print(f'Writing interpolated model to {coeff_name}',
          file=log)
    dask.compute(writes)


    print("All done here.", file=log)



import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from pfb.utils.misc import to4d
import dask.array as da
from dask import delayed
from datetime import datetime
from casacore.quanta import quantity


def data_from_header(hdr, axis=3):
    npix = hdr['NAXIS' + str(axis)]
    refpix = hdr['CRPIX' + str(axis)]
    delta = hdr['CDELT' + str(axis)]
    ref_val = hdr['CRVAL' + str(axis)]
    return ref_val + np.arange(1 - refpix, 1 + npix - refpix) * delta, ref_val


def load_fits(name, dtype=np.float32):
    data = fits.getdata(name)
    data = np.transpose(to4d(data)[:, :, ::-1], axes=(0, 1, 3, 2))
    return np.require(data, dtype=dtype, requirements='C')


def save_fits(data, name, hdr, overwrite=True, dtype=np.float32):
    hdu = fits.PrimaryHDU(header=hdr)
    data = np.transpose(to4d(data), axes=(0, 1, 3, 2))[:, :, ::-1]
    hdu.data = np.require(data, dtype=dtype, requirements='F')
    hdu.writeto(name, overwrite=overwrite)
    return np.ones((1,), dtype=bool)  # so we can use map_blocks


def set_wcs(cell_x, cell_y, nx, ny, radec, freq,
            unit='Jy/beam', GuassPar=None, unix_time=None):
    """
    cell_x/y - cell sizes in degrees
    nx/y - number of x and y pixels
    radec - right ascention and declination in radians
    freq - frequencies in Hz
    """

    w = WCS(naxis=4)
    w.wcs.ctype = ['RA---SIN', 'DEC--SIN', 'FREQ', 'STOKES']
    w.wcs.cdelt[0] = -cell_x
    w.wcs.cdelt[1] = cell_y
    w.wcs.cdelt[3] = 1
    w.wcs.cunit[0] = 'deg'
    w.wcs.cunit[1] = 'deg'
    w.wcs.cunit[2] = 'Hz'
    if np.size(freq) > 1:
        ref_freq = freq[0]
    else:
        ref_freq = freq
    w.wcs.crval = [radec[0]*180.0/np.pi, radec[1]*180.0/np.pi, ref_freq, 1]
    # LB - y axis treated differently because of stupid fits convention?
    w.wcs.crpix = [1 + nx//2, ny//2, 1, 1]

    if np.size(freq) > 1:
        w.wcs.crval[2] = freq[0]
        df = freq[1]-freq[0]
        w.wcs.cdelt[2] = df
        fmean = np.mean(freq)
    else:
        if isinstance(freq, np.ndarray):
            fmean = freq[0]
        else:
            fmean = freq

    header = w.to_header()
    header['RESTFRQ'] = fmean
    header['ORIGIN'] = 'pfb-clean'
    header['BTYPE'] = 'Intensity'
    header['BUNIT'] = unit
    header['SPECSYS'] = 'TOPOCENT'
    if unix_time is not None:
        header['UTC_TIME'] = datetime.utcfromtimestamp(unix_time).strftime('%Y-%m-%d %H:%M:%S')

    return header


def compare_headers(hdr1, hdr2):
    '''
    utility function to ensure that WCS's are compatible
    'NAXIS1', 'NAXIS2', 'NAXIS3', 'NAXIS4',
    '''
    keys = ['CTYPE1', 'CTYPE2', 'CTYPE3', 'CTYPE4',
            'CDELT1', 'CDELT2', 'CDELT3', 'CDELT4',
            'CRPIX1', 'CRPIX2', 'CRPIX3', 'CRPIX4',
            'CUNIT1', 'CUNIT2', 'CUNIT3']
    for key in keys:
        try:
            assert hdr1[key] == hdr2[key]
        except BaseException:
            raise ValueError("Headers do not match on key %s. " % key, hdr1[key], hdr2[key])


def add_beampars(hdr, GaussPar, GaussPars=None):
    """
    Add beam keywords to header.
    GaussPar - MFS beam pars
    GaussPars - beam pars for cube
    """
    if len(GaussPar) == 3:
        hdr['BMAJ'] = GaussPar[0]
        hdr['BMIN'] = GaussPar[1]
        hdr['BPA'] = GaussPar[2]
    elif len(GaussPar) == 1:
        hdr['BMAJ'] = GaussPar[0][0]
        hdr['BMIN'] = GaussPar[0][1]
        hdr['BPA'] = GaussPar[0][2]
    else:
        raise ValueError('Invalid value for GaussPar')

    if GaussPars is not None:
        for i in range(len(GaussPars)):
            hdr['BMAJ' + str(i+1)] = GaussPars[i][0]
            hdr['BMIN' + str(i+1)] = GaussPars[i][1]
            hdr['PA' + str(i+1)] = GaussPars[i][2]

    return hdr


def set_header_info(mhdr, ref_freq, freq_axis, args, beampars):
    hdr_keys = ['SIMPLE', 'BITPIX', 'NAXIS', 'NAXIS1', 'NAXIS2', 'NAXIS3',
                'NAXIS4', 'CTYPE1', 'CTYPE2', 'CTYPE3', 'CTYPE4', 'CRPIX1',
                'CRPIX2', 'CRPIX3', 'CRPIX4', 'CRVAL1', 'CRVAL2', 'CRVAL3',
                'CRVAL4', 'CDELT1', 'CDELT2', 'CDELT3', 'CDELT4']

    new_hdr = {}
    for key in hdr_keys:
        new_hdr[key] = mhdr[key]

    if freq_axis == 3:
        new_hdr["NAXIS3"] = 1
        new_hdr["CRVAL3"] = ref_freq
    elif freq_axis == 4:
        new_hdr["NAXIS4"] = 1
        new_hdr["CRVAL4"] = ref_freq

    new_hdr['BMAJ'] = beampars[0]
    new_hdr['BMIN'] = beampars[1]
    new_hdr['BPA'] = beampars[2]

    new_hdr = fits.Header(new_hdr)

    return new_hdr


def normwsum(data, wsum):
    if wsum > 0:
        return data / wsum
    else:
        return data


def dds2fits(dds, column, outname, norm_wsum=True, otype=np.float32):
    imsout = []
    basename = outname + '_' + column.lower()
    for ds in dds:
        t = ds.timeid
        b = ds.bandid
        name = basename + f'_time{t:04d}_band{b:04d}.fits'
        data = ds.get(column).data
        if norm_wsum:
            data = da.map_blocks(normwsum,
                                 data,
                                 ds.WSUM.data[0],
                                 chunks=data.chunks)
            unit = 'Jy/beam'
        else:
            unit = 'Jy/pixel'
        radec = (ds.ra, ds.dec)
        cell_deg = np.rad2deg(ds.cell_rad)
        nx, ny = ds.get(column).shape
        unix_time = quantity(f'{ds.time_out}s').to_unix_time()
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, radec, ds.freq_out,
                      unit=unit, unix_time=unix_time)
        imout = da.map_blocks(save_fits,
                             data,
                             name,
                             hdr,
                             overwrite=True,
                             chunks=(1,),
                             drop_axis=tuple(np.arange(1,len(data.shape))),
                             meta=np.empty((0,), dtype=bool))
        imsout.append(imout)
    return imsout


def dds2fits_mfs(dds, column, outname, norm_wsum=True, otype=np.float32):
    times_out = []
    freqs = []
    for ds in dds:
        times_out.append(ds.time_out)
        freqs.append(ds.freq_out)
    times_out = np.unique(np.array(times_out))
    ntimes_out = times_out.size
    freq_out = np.mean(np.unique(np.array(freqs)))
    basename = outname + '_' + column.lower()
    imsout = []
    nx, ny = dds[0].get(column).shape
    datas = [da.zeros((nx, ny), chunks=(-1, -1),
                dtype=otype) for _ in range(ntimes_out)]
    wsums = [da.zeros(1) for _ in range(ntimes_out)]
    counts = [da.zeros(1) for _ in range(ntimes_out)]
    radecs = [[] for _ in range(ntimes_out)]
    cell_deg = np.rad2deg(dds[0].cell_rad)
    for ds in dds:
        t = ds.timeid
        datas[t] += ds.get(column).data
        wsums[t] += ds.WSUM.data[0]
        counts[t] += 1
        radecs[t] = (ds.ra, ds.dec)
    for t in range(ntimes_out):
        name = basename + f'_time{t:04d}_mfs.fits'
        if norm_wsum:
            data = da.map_blocks(normwsum,
                                 datas[t],
                                 wsums[t],
                                 chunks=datas[t].chunks)
            unit = 'Jy/beam'
        else:
            data = da.map_blocks(normwsum,
                                 datas[t],
                                 counts[t],
                                 chunks=datas[t].chunks)
            unit = 'Jy/pixel'
        unix_time = quantity(f'{times_out[t]}s').to_unix_time()
        hdr = set_wcs(cell_deg, cell_deg, nx, ny, radecs[t], freq_out,
                      unit=unit, unix_time=unix_time)
        imout = da.map_blocks(save_fits,
                             data,
                             name,
                             hdr,
                             overwrite=True,
                             chunks=(1,),
                             drop_axis=tuple(np.arange(1,len(data.shape))),
                             meta=np.empty((0,), dtype=bool))
        imsout.append(imout)

    return imsout



import numba

from pfb.wavelets.modes import Modes


@numba.njit(nogil=True, cache=True)
def upsampling_convolution_valid_sf(input, filter, output, mode):
    if mode is not Modes.zeropad:
        raise NotImplementedError("Only zeropad mode supported")

    N = input.shape[0]
    F = filter.shape[0]
    O = output.shape[0]

    if F % 2:
        raise ValueError("even filter required for upsampling")

    if N < (F // 2):
        raise ValueError("input.shape[0] < (filter.shape[0] // 2)")

    o = 0
    i = (F // 2) - 1

    while i < N:
        sum_even = input.dtype.type(0)
        sum_odd = input.dtype.type(0)
        j = 0

        while j < F // 2:
            sum_even += filter[j * 2] * input[i - j]
            sum_odd += filter[j * 2 + 1] * input[i - j]

            j += 1

        output[o] += sum_even
        output[o + 1] += sum_odd

        i += 1
        o += 2


@numba.njit(nogil=True, cache=True)
def downsampling_convolution(input, output, filter, mode, step):
    if mode is not Modes.zeropad:
        raise NotImplementedError("Only zeropad mode supported")

    i = step - 1
    o = 0
    N = input.shape[0]
    F = filter.shape[0]

    if mode is Modes.smooth and N < 2:
        mode = Modes.constant_edge

    # left boundary overhang
    while i < F and i < N:
        fsum = input.dtype.type(0)
        j = 0

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if input equal or wider than filter: N >= F)
    while i < N:
        fsum = input.dtype.type(0)
        j = 0

        while j < F:
            fsum += input[i - j] * filter[j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # center (if filter is wider than input: F > N)
    while i < F:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j <= i:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    # right boundary overhang
    while i < N + F - 1:
        fsum = input.dtype.type(0)
        j = 0
        j = i - N + 1

        while j < F:
            fsum += filter[j] * input[i - j]
            j += 1

        output[o] = fsum

        i += step
        o += 1

    if o != output.shape[0]:
        raise ValueError("Output mismatch")



import numpy as np


def prox_21(v, sigma, weight=None, axis=0):
    """
    Computes weighted version of

    prox_{sigma * || . ||_{21}}(v)

    Assumed that v has shape (nband, nbasis, ntot) where

    nband   - number of imaging bands
    nbasis  - number of orthogonal bases
    ntot    - total number of coefficients for each basis (must be equal)
    """
    l2_norm = np.linalg.norm(v, axis=axis)  # drops axis
    l2_soft = np.maximum(l2_norm - sigma * weight, 0.0)  # norm positive
    mask = l2_norm != 0
    ratio = np.zeros(mask.shape, dtype=v.dtype)
    ratio[mask] = l2_soft[mask] / l2_norm[mask]
    return v * np.expand_dims(ratio, axis=axis)  # restores axis
